#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
obj: MMGen native classes
"""

import unicodedata

from .objmethods import MMGenObject, Hilite, HiliteStr, InitErrors

def get_obj(objname, *args, **kwargs):
	"""
	Wrapper for data objects
	- If the object throws an exception on instantiation, return False, otherwise return the object.
	- If silent is True, suppress display of the exception.
	- If return_bool is True, return True instead of the object.
	Only keyword args are accepted.
	"""
	assert not args, 'get_obj_chk1'

	silent, return_bool = (False, False)
	if 'silent' in kwargs:
		silent = kwargs['silent']
		del kwargs['silent']
	if 'return_bool' in kwargs:
		return_bool = kwargs['return_bool']
		del kwargs['return_bool']

	try:
		ret = objname(**kwargs)
	except Exception as e:
		if not silent:
			from .util import rmsg
			rmsg(f'{e!s}')
		return False
	else:
		return True if return_bool else ret

# dict that keeps a list of keys for efficient lookup by index
class IndexedDict(dict):

	def __init__(self, *args, **kwargs):
		if args or kwargs:
			self.die('initializing values via constructor')
		self.__keylist = []
		dict.__init__(self, *args, **kwargs)

	def __setitem__(self, key, value):
		if key in self:
			self.die('reassignment to existing key')
		self.__keylist.append(key)
		return dict.__setitem__(self, key, value)

	@property
	def keys(self):
		return self.__keylist

	def key(self, idx):
		return self.__keylist[idx]

	def __delitem__(self, *args):
		self.die('item deletion')

	def move_to_end(self, *args):
		self.die('item moving')

	def clear(self, *args):
		self.die('clearing')

	def update(self, *args):
		self.die('updating')

	def die(self, desc):
		raise NotImplementedError(f'{desc} not implemented for type {type(self).__name__}')

class MMGenList(list, MMGenObject):
	pass

class MMGenDict(dict, MMGenObject):
	pass

class ImmutableAttr: # Descriptor
	"""
	For attributes that are always present in the data instance
	Reassignment and deletion forbidden
	"""
	ok_dtypes = (type, type(None), type(lambda:0))

	def __init__(self, dtype, *, typeconv=True, set_none_ok=False, include_proto=False):
		self.set_none_ok = set_none_ok
		self.typeconv = typeconv

		assert isinstance(dtype, self.ok_dtypes), 'ImmutableAttr_check1'

		if set_none_ok:
			assert typeconv and not isinstance(dtype, str), 'ImmutableAttr_check2'

		if typeconv:
			# convert this attribute's type
			if set_none_ok:
				self.conv = lambda instance, value: None if value is None else dtype(value)
			elif include_proto:
				self.conv = lambda instance, value: dtype(instance.proto, value)
			else:
				self.conv = lambda instance, value: dtype(value)
		else:
			# check this attribute's type
			def assign_with_check(instance, value):
				if type(value) is dtype:
					return value
				raise TypeError('Attribute {!r} of {} instance must be of type {}, not {}'.format(
					self.name,
					type(instance).__name__,
					dtype,
					type(value)))
			self.conv = assign_with_check

	def __set_name__(self, owner, name):
		self.name = name

	def __get__(self, instance, owner):
		return instance.__dict__[self.name]

	def setattr_condition(self, instance):
		'forbid all reassignment'
		return not self.name in instance.__dict__

	def __set__(self, instance, value):
		if not self.setattr_condition(instance):
			raise AttributeError(f'Attribute {self.name!r} of {type(instance)} instance cannot be reassigned')
		instance.__dict__[self.name] = self.conv(instance, value)

	def __delete__(self, instance):
		raise AttributeError(
			f'Attribute {self.name!r} of {type(instance).__name__} instance cannot be deleted')

class ListItemAttr(ImmutableAttr):
	"""
	For attributes that might not be present in the data instance
	Reassignment or deletion allowed if specified
	"""
	def __init__(self, dtype, *, typeconv=True, include_proto=False, reassign_ok=False, delete_ok=False):
		self.reassign_ok = reassign_ok
		self.delete_ok = delete_ok
		ImmutableAttr.__init__(self, dtype, typeconv=typeconv, include_proto=include_proto)

	def __get__(self, instance, owner):
		"return None if attribute doesn't exist"
		try:
			return instance.__dict__[self.name]
		except:
			return None

	def setattr_condition(self, instance):
		return getattr(instance, self.name) is None or self.reassign_ok

	def __delete__(self, instance):
		if self.delete_ok:
			if self.name in instance.__dict__:
				del instance.__dict__[self.name]
		else:
			ImmutableAttr.__delete__(self, instance)

class MMGenListItem(MMGenObject):
	valid_attrs = set()
	invalid_attrs = {
		'pfmt',
		'pmsg',
		'pdie',
		'pexit',
		'valid_attrs',
		'invalid_attrs',
		'immutable_attr_init_check',
	}

	def __init__(self, *args, **kwargs):
		# generate valid_attrs, or use the class valid_attrs if set
		self.__dict__['valid_attrs'] = self.valid_attrs or (
				{e for e in dir(self) if e[0] != '_'}
				- MMGenListItem.invalid_attrs
				- self.invalid_attrs
			)

		if args:
			raise ValueError(f'Non-keyword args not allowed in {type(self).__name__!r} constructor')

		for k, v in kwargs.items():
			if v is not None:
				setattr(self, k, v)

		# Require all immutables to be initialized.  Check performed only when testing.
		self.immutable_attr_init_check()

	# allow only valid attributes to be set
	def __setattr__(self, name, value):
		if name not in self.valid_attrs:
			raise AttributeError(f'{name!r}: no such attribute in class {type(self)}')
		return object.__setattr__(self, name, value)

	def _asdict(self):
		return dict((k, v) for k, v in self.__dict__.items() if k in self.valid_attrs)

class MMGenRange(tuple, InitErrors, MMGenObject):

	min_idx = None
	max_idx = None

	def __new__(cls, *args):
		try:
			if len(args) == 1:
				s = args[0]
				if isinstance(s, cls):
					return s
				assert isinstance(s, str), 'not a string or string subclass'
				ss = s.split('-', 1)
				first = int(ss[0])
				last = int(ss.pop())
			else:
				s = repr(args) # needed if exception occurs
				assert len(args) == 2, 'one format string arg or two start, stop args required'
				first, last = args
			assert first <= last, 'start of range greater than end of range'
			if cls.min_idx is not None:
				assert first >= cls.min_idx, f'start of range < {cls.min_idx:,}'
			if cls.max_idx is not None:
				assert last <= cls.max_idx, f'end of range > {cls.max_idx:,}'
			return tuple.__new__(cls, (first, last))
		except Exception as e:
			return cls.init_fail(e, s)

	@property
	def first(self):
		return self[0]

	@property
	def last(self):
		return self[1]

	def iterate(self):
		return range(self[0], self[1]+1)

	@property
	def items(self):
		return list(self.iterate())

class Int(int, Hilite, InitErrors):
	min_val = None
	max_val = None
	max_digits = None
	color = 'red'

	def __new__(cls, n, *, base=10):
		if isinstance(n, cls):
			return n
		try:
			me = int.__new__(cls, str(n), base)
			if cls.min_val is not None:
				assert me >= cls.min_val, f'is less than cls.min_val ({cls.min_val})'
			if cls.max_val is not None:
				assert me <= cls.max_val, f'is greater than cls.max_val ({cls.max_val})'
			if cls.max_digits is not None:
				assert len(str(me)) <= cls.max_digits, f'has more than {cls.max_digits} digits'
			return me
		except Exception as e:
			return cls.init_fail(e, n)

	@classmethod
	def fmtc(cls, s, width, /, *, color=False):
		return super().fmtc(str(s), width, color=color)

	def fmt(self, width, /, *, color=False):
		return super().fmtc(str(self), width, color=color)

	def hl(self, **kwargs):
		return super().colorize(str(self), **kwargs)

class NonNegativeInt(Int):
	min_val = 0

class MMGenIdx(Int):
	min_val = 1

class ETHNonce(Int):
	min_val = 0

class Str(HiliteStr):
	pass

class HexStr(HiliteStr, InitErrors):
	color = 'red'
	width = None
	hexcase = 'lower'
	trunc_ok = False
	def __new__(cls, s, *, case=None):
		if isinstance(s, cls):
			return s
		if case is None:
			case = cls.hexcase
		from .util import hexdigits_lc, hexdigits_uc
		try:
			assert isinstance(s, str), 'not a string or string subclass'
			assert case in ('upper', 'lower'), f'{case!r} incorrect case specifier'
			assert set(s) <= set(hexdigits_lc if case == 'lower' else hexdigits_uc), (
				f'not {case}case hexadecimal symbols')
			assert not len(s) % 2, 'odd-length string'
			if cls.width:
				assert len(s) == cls.width, f'Value is not {cls.width} characters wide'
			return str.__new__(cls, s)
		except Exception as e:
			return cls.init_fail(e, s)

	def truncate(self, width, /, *, color=True):
		return self.colorize(
			self if width >= self.width else self[:width-2] + '..',
			color = color)

class CoinTxID(HexStr):
	color, width, hexcase = ('purple', 64, 'lower')

def is_coin_txid(s):
	return get_obj(CoinTxID, s=s, silent=True, return_bool=True)

class WalletPassword(HexStr):
	color, width, hexcase = ('blue', 32, 'lower')

class MMGenTxID(HexStr):
	color, width, hexcase = ('red', 6, 'upper')

class MMGenLabel(HiliteStr, InitErrors):
	color = 'pink'
	allowed = []
	forbidden = []
	first_char = []
	max_len = 0
	min_len = 0
	max_screen_width = 0 # if != 0, overrides max_len
	desc = 'label'
	def __new__(cls, s):
		if isinstance(s, cls):
			return s
		try:
			s = s.strip()
			if not cls.allowed:
				for ch in s:
					# Allow:    (L)etter, (N)umber, (P)unctuation, (S)ymbol, (Z)space
					# Disallow: (C)ontrol, (M)combining
					# Combining characters create width formatting issues, so disallow them for now
					if unicodedata.category(ch)[0] in ('C', 'M'):
						raise ValueError(
							'{!a}: {} characters not allowed'.format(
								ch,
								{'C':'control', 'M':'combining'}[unicodedata.category(ch)[0]]))

			me = str.__new__(cls, s)

			if cls.max_screen_width:
				me.screen_width = len(s) + len([1 for ch in s if unicodedata.east_asian_width(ch) in ('F', 'W')])
				assert me.screen_width <= cls.max_screen_width, f'too wide (>{cls.max_screen_width} screen width)'
			else:
				assert len(s) <= cls.max_len, f'too long (>{cls.max_len} symbols)'

			assert len(s) >= cls.min_len, f'too short (<{cls.min_len} symbols)'

			if cls.first_char and s and not s[0] in cls.first_char:
				raise ValueError('first character not in set ' + ' '.join(cls.first_char))

			if cls.allowed and not set(list(s)).issubset(set(cls.allowed)):
				raise ValueError('contains symbols not in set: ' + ' '.join(cls.allowed))

			if cls.forbidden and any(ch in s for ch in cls.forbidden):
				raise ValueError('contains one of these forbidden symbols: ' + ' '.join(cls.forbidden))

			return me
		except Exception as e:
			return cls.init_fail(e, s)

class MMGenWalletLabel(MMGenLabel):
	max_len = 48
	desc = 'wallet label'

class TrackingWalletName(MMGenLabel):
	max_len = 40
	desc = 'tracking wallet name'
	allowed = 'abcdefghijklmnopqrstuvwxyz0123456789-_'
	first_char = 'abcdefghijklmnopqrstuvwxyz0123456789'

class TwComment(MMGenLabel):
	max_screen_width = 80
	desc = 'tracking wallet comment'
	exc = 'BadTwComment'

class MMGenTxComment(MMGenLabel):
	max_len = 72
	desc = 'transaction label'

class MMGenPWIDString(MMGenLabel):
	max_len = 256
	min_len = 1
	desc = 'password ID string'
	forbidden = list(' :/\\')
	trunc_ok = False

class Hostname(MMGenLabel):
	max_len = 256
	color = 'pink'



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# Project source code repository: https://github.com/mmgen/mmgen-wallet
# Licensed according to the terms of GPL Version 3.  See LICENSE for details.

"""
test.cmdtest_d.cfgfile: CfgFile tests for the MMGen cmdtest.py test suite
"""

import sys, os, time, shutil

from mmgen.color import yellow
from mmgen.cfgfile import CfgFileSampleSys, CfgFileSampleUsr, cfg_file_sample

from ..include.common import read_from_file, write_to_file, imsg
from .base import CmdTestBase

class CmdTestCfgFile(CmdTestBase):
	'CfgFile API'
	networks = ('btc',)
	tmpdir_nums = [40]
	base_passthru_opts = ()
	color = True

	cmd_group = (
		('sysfile',                  (40, 'init with system cfg sample file in place', [])),
		('opts_data_sets1',          (40, 'opts_data["sets"] opt set in environment', [])),
		('opts_data_sets2',          (40, 'opts_data["sets"] opt set in cfg_file', [])),
		('no_metadata_sample',       (40, 'init with unversioned cfg sample file', [])),
		('altered_sample',           (40, 'init with user-modified cfg sample file', [])),
		('old_sample',               (40, 'init with old v2 cfg sample file', [])),
		('old_sample_bad_var',       (40, 'init with old v2 cfg sample file and bad variable in mmgen.cfg', [])),
		('autoset_opts',             (40, 'setting autoset opts', [])),
		('autoset_opts_cmdline',     (40, 'setting autoset opts (override on cmdline)', [])),
		('autoset_opts_bad',         (40, 'setting autoset opts (bad value in cfg file)', [])),
		('autoset_opts_bad_cmdline', (40, 'setting autoset opts (bad param on cmdline)', [])),
		('coin_specific_vars',       (40, 'setting coin-specific vars', [])),
		('chain_names',              (40, 'setting chain names', [])),
		('mnemonic_entry_modes',     (40, 'setting mnemonic entry modes', [])),
		('opt_override1',            (40, 'cfg file opts not overridden', [])),
		('opt_override2',            (40, 'negative cmdline opts overriding cfg file opts', [])),
	)

	def __init__(self, cfg, trunner, cfgs, spawn):
		CmdTestBase.__init__(self, cfg, trunner, cfgs, spawn)
		self.spawn_env['MMGEN_TEST_SUITE_CFGTEST'] = '1'

	def read_from_cfgfile(self, loc):
		return read_from_file(self.path(loc))

	def write_to_cfgfile(self, loc, data, verbose=False):
		write_to_file(self.path(loc), '\n'.join(data) + '\n')
		if verbose:
			imsg(yellow(f'Wrote cfg file: {data!r}'))

	def spawn_test(self, opts=[], args=[], extra_desc='', pexpect_spawn=None, exit_val=None):
		return self.spawn(
			'test/misc/cfg.py',
			[f'--data-dir={self.path("data_dir")}'] + opts + args,
			cmd_dir       = '.',
			extra_desc    = extra_desc,
			pexpect_spawn = pexpect_spawn,
			exit_val      = exit_val)

	def path(self, id_str):
		return {
			'ref':         'test/ref/mmgen.cfg',
			'data_dir':    '{}/data_dir'.format(self.tmpdir),
			'shared_data': '{}/data_dir/{}'.format(self.tmpdir, CfgFileSampleSys.test_fn_subdir),
			'usr':         '{}/data_dir/mmgen.cfg'.format(self.tmpdir),
			'sys':         '{}/data_dir/{}/mmgen.cfg'.format(self.tmpdir, CfgFileSampleSys.test_fn_subdir),
			'sample':      '{}/data_dir/mmgen.cfg.sample'.format(os.path.abspath(self.tmpdir)),
		}[id_str]

	def copy_sys_sample(self):
		os.makedirs(self.path('shared_data'), exist_ok=True)
		shutil.copy2(self.path('ref'), self.path('sys'))

	def sysfile(self):
		self.copy_sys_sample()
		t = self.spawn_test()
		t.read()
		u = self.read_from_cfgfile('usr')
		S = self.read_from_cfgfile('sys')
		assert u[-1] == '\n', u
		assert u.replace('\r\n', '\n') == S, 'u != S'
		self.check_replaced_sample()
		return t

	def check_replaced_sample(self):
		s = self.read_from_cfgfile('sample')
		S = self.read_from_cfgfile('sys')
		assert s[-1] == '\n', s
		assert S.splitlines() == s.splitlines()[:-1], 'sys != sample[:-1]'

	def bad_sample(self, s, e):
		write_to_file(self.path('sample'), s)
		t = self.spawn_test()
		t.expect(e)
		t.read()
		self.check_replaced_sample()
		return t

	def opts_data_sets1(self): # no_license (in env) sets grokify
		self.write_to_cfgfile('usr', ['scroll true'])
		t = self.spawn_test(args=['print_cfg', 'no_license', 'foobleize', 'grokify', 'scroll'])
		t.expect('foobleize: None')
		t.expect('grokify: True')
		return t

	def opts_data_sets2(self): # autosign (in cfg file) sets foobleize
		self.write_to_cfgfile('usr', ['autosign true'])
		t = self.spawn_test(args=['print_cfg', 'no_license', 'autosign', 'foobleize', 'grokify'])
		t.expect('foobleize: True')
		t.expect('grokify: True')
		return t

	def no_metadata_sample(self):
		self.copy_sys_sample()
		S = self.read_from_cfgfile('sys')
		e = CfgFileSampleUsr.out_of_date_fs.format(self.path('sample'))
		return self.bad_sample(S, e)

	def altered_sample(self):
		s = '\n'.join(self.read_from_cfgfile('sample').splitlines()[1:]) + '\n'
		e = CfgFileSampleUsr.altered_by_user_fs.format(self.path('sample'))
		return self.bad_sample(s, e)

	def old_sample_common(self, old_set=False, args=[], pexpect_spawn=False):
		d = (
			self.read_from_cfgfile('sys').replace('monero_', 'zcash_').splitlines()
			+ ['', '# Uncomment to make foo true:', '# foo true']
			+ ['', '# Uncomment to make bar false:', '# bar false']
		)
		self.write_to_cfgfile('sample', d + cfg_file_sample.cls_make_metadata(d))

		t = self.spawn_test(args=args, pexpect_spawn=pexpect_spawn, exit_val=1 if old_set else None)

		t.expect('options have changed')
		for s in ('have been added', 'monero_', 'have been removed', 'zcash_', 'foo', 'bar'):
			t.expect(s)

		if old_set:
			for s in ('must be deleted', 'bar', 'foo'):
				t.expect(s)

		cp = CfgFileSampleUsr.details_confirm_prompt + ' (y/N): '

		t.expect(cp, 'y')

		for s in ('CHANGES', 'Removed', '# zcash_', '# foo', '# bar', 'Added', '# monero_'):
			t.expect(s)

		if t.pexpect_spawn: # view and exit pager
			time.sleep(1 if self.cfg.exact_output else t.send_delay)
			t.send('q')

		t.expect(cp, 'n')

		if old_set:
			t.expect('unrecognized option')

		if args == ['parse_test']:
			t.expect('parsed chunks: 29')
			t.expect('usr cfg: testnet=true rpc_password=passwOrd')

		if not old_set:
			self.check_replaced_sample()

		return t

	def old_sample(self):
		self.write_to_cfgfile('usr', ['testnet true', 'rpc_password passwOrd'])
		return self.old_sample_common(args=['parse_test'])

	def old_sample_bad_var(self):
		self.write_to_cfgfile('usr', ['foo true', 'bar false'])
		t = self.old_sample_common(
			old_set       = True,
			pexpect_spawn = not sys.platform == 'win32')
		t.expect('unrecognized option')
		return t

	def _autoset_opts(self, args=[], text='rpc_backend aiohttp', exit_val=None):
		self.write_to_cfgfile('usr', [text], verbose=True)
		return self.spawn_test(args=args, exit_val=exit_val)

	def autoset_opts(self):
		return self._autoset_opts(args=['autoset_opts'])

	def autoset_opts_cmdline(self):
		return self._autoset_opts(args=['--rpc-backend=curl', 'autoset_opts_cmdline'])

	def _autoset_opts_bad(self, expect, kwargs):
		t = self._autoset_opts(exit_val=1, **kwargs)
		t.expect(expect)
		return t

	def autoset_opts_bad(self):
		return self._autoset_opts_bad('not unique substring', {'text':'rpc_backend foo'})

	def autoset_opts_bad_cmdline(self):
		return self._autoset_opts_bad('not unique substring', {'args':['--rpc-backend=foo']})

	def coin_specific_vars(self):
		"""
		ensure that derived classes explicitly set these variables
		"""

		if self.cfg.no_altcoin:
			return 'skip'

		d = [
			'btc_max_tx_fee 1.2345',
			'eth_max_tx_fee 5.4321',
			'btc_ignore_daemon_version true',
			'eth_ignore_daemon_version true'
		]
		self.write_to_cfgfile('usr', d, verbose=True)

		for coin, res1_chk, res2_chk, res2_chk_eq in (
			('BTC', 'True',  '1.2345', True),
			('LTC', 'None',  '1.2345', False),
			('BCH', 'None',  '1.2345', False),
			('ETH', 'True',  '5.4321', True),
			('ETC', 'None',  '5.4321', False)
		):
			t = self.spawn_test(
				args = [
					f'--coin={coin}',
					'coin_specific_vars',
					'ignore_daemon_version',
					'max_tx_fee'
				],
				extra_desc=f'({coin})')
			res1 = t.expect_getend('ignore_daemon_version: ')
			res2 = t.expect_getend('max_tx_fee: ')
			assert res1 == res1_chk, f'{res1} != {res1_chk}'
			if res2_chk_eq:
				assert res2 == res2_chk, f'{res2} != {res2_chk}'
			else:
				assert res2 != res2_chk, f'{res2} == {res2_chk}'
			t.read()
			t.ok()

		t.skip_ok = True
		return t

	def mnemonic_entry_modes(self):

		def run(modes_chk):
			t = self.spawn_test(args=['mnemonic_entry_modes'])
			modes = t.expect_getend('mnemonic_entry_modes: ')
			assert modes_chk == modes, f'{modes_chk} != {modes}'
			return t

		self.write_to_cfgfile('usr', ['mnemonic_entry_modes mmgen:full bip39:short'], verbose=True)

		t = run("{'mmgen': 'full', 'bip39': 'short'}")
		# check that set_dfl_entry_mode() set the mode correctly:
		t.expect('mmgen: full')
		t.expect('bip39: short')
		return t

	def chain_names(self):

		if self.cfg.no_altcoin:
			return 'skip'

		def run(chk, testnet):
			for coin, chain_chk in (('ETH', chk), ('ETC', None)):
				t = self.spawn_test(
					args = [f'--coin={coin}', f'--testnet={(0, 1)[testnet]}', 'coin_specific_vars', 'chain_names'],
					extra_desc = f'({coin} testnet={testnet!r:5} chain_names={chain_chk})')
				chain = t.expect_getend('chain_names: ')
				if chain_chk:
					assert chain == chain_chk, f'{chain} != {chain_chk}'
				else:
					assert chain != chain_chk, f'{chain} == {chain_chk}'
				t.read()
				t.ok()
			return t

		self.write_to_cfgfile('usr', ['eth_mainnet_chain_names istanbul constantinople'], verbose=True)

		t = run("['istanbul', 'constantinople']", False)
		t = run(None, True)

		self.write_to_cfgfile('usr', ['eth_testnet_chain_names rinkeby'], verbose=True)

		t = run(None, False)
		t = run("['rinkeby']", True)

		t.skip_ok = True
		return t

	def opt_override1(self):
		self.write_to_cfgfile('usr', ['no_license true', 'scroll true'])
		t = self.spawn_test(
			args = ['print_cfg', 'scroll', 'no_license'])
		t.expect('scroll: True')
		t.expect('no_license: True')
		return t

	def opt_override2(self):
		self.write_to_cfgfile('usr', ['no_license true', 'scroll true'])
		t = self.spawn_test(
			args = ['print_cfg', 'scroll', 'no_license'],
			opts = ['--no-scrol', '--lic'])
		t.expect('scroll: False')
		t.expect('no_license: False')
		return t



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_d.misc: Miscellaneous test groups for the cmdtest.py test suite
"""

import sys, re

from mmgen.util import die

from ..include.common import start_test_daemons, stop_test_daemons, imsg
from .include.common import get_file_with_ext, dfl_words_file
from .base import CmdTestBase
from .main import CmdTestMain

class CmdTestDev(CmdTestBase):
	'developer scripts'
	networks = ('btc',)
	cmd_group = (
		('compute_file_chksum', 'scripts/compute-file-chksum.py'),
		('create_bip_hd_chain_params', 'scripts/create-bip-hd-chain-params.py'),
	)
	tmpdir_nums = [99]
	color = True

	def _spawn(self, script, args):
		return self.spawn(script, args, cmd_dir='.', no_exec_wrapper=True)

	def compute_file_chksum(self):
		t = self._spawn('scripts/compute-file-chksum.py', ['test/ref/25EFA3[2.34].testnet.rawtx'])
		t.expect('3df942')
		return t

	def create_bip_hd_chain_params(self):
		t = self._spawn('scripts/create-bip-hd-chain-params.py', ['test/ref/altcoin/slip44-mini.json'])
		t.expect('[defaults]')
		t.expect(r"secp.*0488ade4.*0488b21e.*0'\/0\/0", regex=True)
		t.expect('[bip-44]')
		t.expect('[bip-49]')
		t.match_expect_list(
			['0', 'BTC', 'x', 'm', 'P2SH', '049d7878', '049d7cb2', '80', '05', 'x', 'Bitcoin', '1'])
		return t

class CmdTestMisc(CmdTestBase):
	'miscellaneous tests (RPC backends, xmrwallet_txview, term)'
	networks = ('btc',)
	tmpdir_nums = [99]
	passthru_opts = ('daemon_data_dir', 'rpc_port')
	cmd_group = (
		('rpc_backends',         'RPC backends'),
		('bch_txview_legacy1',   '‘mmgen-tool --coin=bch --cashaddr=0 txview terse=0’'),
		('bch_txview_legacy2',   '‘mmgen-tool --coin=bch --cashaddr=0 txview terse=1’'),
		('bch_txview_cashaddr1', '‘mmgen-tool --coin=bch --cashaddr=1 txview terse=0’'),
		('bch_txview_cashaddr2', '‘mmgen-tool --coin=bch --cashaddr=1 txview terse=1’'),
		('xmrwallet_txview',     '‘mmgen-xmrwallet txview’'),
		('xmrwallet_txlist',     '‘mmgen-xmrwallet txlist’'),
		('coin_daemon_info',     '‘examples/coin-daemon-info.py’'),
		('examples_bip_hd',      '‘examples/bip_hd.py’'),
		('term_echo',            'term.set("echo")'),
		('term_cleanup',         'term.register_cleanup()'),
	)
	need_daemon = True
	color = True

	def rpc_backends(self):
		backends = self.cfg._autoset_opts['rpc_backend'][1]
		for b in backends:
			t = self.spawn_chk('mmgen-tool', [f'--rpc-backend={b}', 'daemon_version'], extra_desc=f'({b})')
		return t

	def _bch_txview(self, view_pref, terse, expect):
		if self.cfg.no_altcoin:
			return 'skip'
		tx = 'test/ref/bitcoin_cash/895108-BCH[2.65913].rawtx'
		t = self.spawn('mmgen-tool', ['--coin=bch', f'--cashaddr={view_pref}', 'txview', tx, f'terse={terse}'])
		#t = self.spawn('mmgen-tool', ['--coin=bch', '--longhelp'])
		t.expect(expect)
		return t

	def bch_txview_legacy1(self):
		return self._bch_txview(0, 0, '[qzuffa536e0eqfwz3smapckhlw9wge4p5spvx5j7h7]')

	def bch_txview_legacy2(self):
		return self._bch_txview(0, 1, '[qzuffa536e0eqfwz3smapckhlw9wge4p5spvx5j7h7]')

	def bch_txview_cashaddr1(self):
		return self._bch_txview(1, 0, '[1HpynST7vkLn8yNtdrqPfeghexZk4sdB3W]')

	def bch_txview_cashaddr2(self):
		return self._bch_txview(1, 1, '[1HpynST7vkLn8yNtdrqPfeghexZk4sdB3W]')

	def xmrwallet_txview(self, op='txview'):
		if self.cfg.no_altcoin:
			return 'skip'
		files = get_file_with_ext('test/ref/monero', 'tx', no_dot=True, delete=False, return_list=True)
		t = self.spawn('mmgen-xmrwallet', [op] + files)
		res = t.read(strip_color=True)
		if op == 'txview':
			for s in (
				'Amount:    0.74 XMR',
				'Dest:      56VQ9M6k',
			):
				assert s in res, f'{s} not in {res}'
		elif op == 'txlist':
			assert re.search('3EBD06-.*D94583-.*8BFA29-', res, re.DOTALL)
		return t

	def xmrwallet_txlist(self):
		return self.xmrwallet_txview(op='txlist')

	def examples_bip_hd(self):
		if self.cfg.no_altcoin:
			return 'skip'
		return self.spawn('examples/bip_hd.py', cmd_dir='.')

	def coin_daemon_info(self):
		if self.cfg.no_altcoin:
			coins = ['btc']
		else:
			coins = ['btc', 'ltc', 'eth']
			start_test_daemons('ltc', 'eth')
		t = self.spawn('examples/coin-daemon-info.py', coins, cmd_dir='.')
		for coin in coins:
			t.expect(coin.upper() + r'\s+mainnet\s+Up', regex=True)
		if self.cfg.pexpect_spawn:
			t.send('q')
		if not self.cfg.no_altcoin:
			stop_test_daemons('ltc', 'eth', remove_datadir=True)
		return t

	def term_echo(self):

		def test_echo():
			t.expect('echo> ', 'foo\n')
			t.expect('foo')

		def test_noecho():
			t.expect('noecho> ', 'foo\n')
			import pexpect
			try:
				t.expect('foo')
			except pexpect.TIMEOUT:
				imsg('[input not echoed - OK]')
			else:
				die('TestSuiteException', 'Terminal echoed in noecho mode!')
			t.send('x')

		if self.skip_for_win('no termios support') or self.skip_for_mac('termios.ECHO issues'):
			return 'skip'

		t = self.spawn('test/misc/term_ni.py', ['echo'], cmd_dir='.', pexpect_spawn=True, timeout=1)
		t.p.logfile = None
		t.p.logfile_read = None if self.tr.quiet else sys.stdout
		t.p.logfile_send = None

		test_noecho()
		test_echo()
		test_noecho()

		return t

	def term_cleanup(self):
		if self.skip_for_win('no termios support'):
			return 'skip'
		return self.spawn('test/misc/term_ni.py', ['cleanup'], cmd_dir='.', pexpect_spawn=True)

class CmdTestOutput(CmdTestBase):
	'screen output'
	networks = ('btc',)
	cmd_group = (
		('output_gr',            (1, 'Greek text', [])),
		('output_ru',            (1, 'Russian text', [])),
		('output_zh',            (1, 'Chinese text', [])),
		('output_jp',            (1, 'Japanese text', [])),
		('oneshot_warning',      (1, 'Oneshot warnings', [])),
		('oneshot_warning_term', (1, 'Oneshot warnings (pexpect_spawn)', []))
	)
	color = True

	def screen_output(self, lang):
		return self.spawn('test/misc/utf8_output.py', [lang], cmd_dir='.')

	def output_gr(self):
		return self.screen_output('gr')
	def output_ru(self):
		return self.screen_output('ru')
	def output_zh(self):
		return self.screen_output('zh')
	def output_jp(self):
		return self.screen_output('jp')

	def oneshot_warning(self, pexpect_spawn=None):
		t = self.spawn('test/misc/oneshot_warning.py', cmd_dir='.', pexpect_spawn=pexpect_spawn)
		nl = '\r\n' if sys.platform == 'win32' or t.pexpect_spawn else '\n'
		for s in (
			f'pw{nl}wg1',
			'foo is experimental',
			'wg2', 'The bar command is dangerous',
			'wg3', 'baz variant alpha',
			'wg4', 'baz variant beta',
			'w1', 'foo variant alpha',
			'w2', 'foo variant beta',
			'w3', 'bar is experimental',
			'pw',
			"passphrase from file 'A'",
			"passphrase from file 'B'",
			f'wg1{nl}wg2{nl}wg3{nl}wg4{nl}w1{nl}w2{nl}w3',
			'pw',
			"passphrase from file 'A'",
			"passphrase from file 'B'",
			f'wg1{nl}wg2{nl}wg3{nl}wg4{nl}w1{nl}w2{nl}w3',
			'bottom',
		):
			t.expect(s)
		return t

	def oneshot_warning_term(self):
		if self.skip_for_win('no pexpect_spawn'):
			return 'skip'
		return self.oneshot_warning(pexpect_spawn=True)

class CmdTestRefTX(CmdTestMain, CmdTestBase):
	'create a reference transaction file (administrative command)'
	segwit_opts_ok = False
	passthru_opts = ('daemon_data_dir', 'rpc_port', 'coin', 'testnet')
	tmpdir_nums = [31, 32, 33, 34]
	need_daemon = True
	cmd_group = (
		('ref_tx_addrgen1', (31, 'address generation (legacy)',     [[[], 1]])),
		('ref_tx_addrgen2', (32, 'address generation (compressed)', [[[], 1]])),
		('ref_tx_addrgen3', (33, 'address generation (segwit)',     [[[], 1]])),
		('ref_tx_addrgen4', (34, 'address generation (bech32)',     [[[], 1]])),
		('ref_tx_txcreate',
			(31, 'transaction creation',
			([['addrs'], 31], [['addrs'], 32], [['addrs'], 33], [['addrs'], 34]))
		),
	)

	def __init__(self, cfg, trunner, cfgs, spawn):
		if cfgs:
			for n in self.tmpdir_nums:
				cfgs[str(n)].update({
					'addr_idx_list': '1-2',
					'segwit': n in (33, 34),
					'dep_generators': {'addrs':'ref_tx_addrgen'+str(n)[-1]}
				})
		CmdTestMain.__init__(self, cfg, trunner, cfgs, spawn)

	def ref_tx_addrgen(self, atype):
		if atype not in self.proto.mmtypes:
			return
		return self.spawn('mmgen-addrgen', ['--outdir='+self.tmpdir, '--type='+atype, dfl_words_file, '1-2'])

	def ref_tx_addrgen1(self):
		return self.ref_tx_addrgen(atype='L')
	def ref_tx_addrgen2(self):
		return self.ref_tx_addrgen(atype='C')
	def ref_tx_addrgen3(self):
		return self.ref_tx_addrgen(atype='S')
	def ref_tx_addrgen4(self):
		return self.ref_tx_addrgen(atype='B')

	def ref_tx_txcreate(self, f1, f2, f3, f4):
		sources = ['31', '32']
		if 'S' in self.proto.mmtypes:
			sources += ['33']
		if 'B' in self.proto.mmtypes:
			sources += ['34']
		return self.txcreate_common(
				addrs_per_wallet = 2,
				sources          = sources,
				add_args         = ['--locktime=1320969600'],
				do_label         = True)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.mmhex: MMGen hexadecimal file wallet class
"""

from ..util import make_chksum_6, split_into_cols
from ..seed import Seed
from ..util import msg, is_chksum_6, is_hex_str
from .unenc import wallet

class wallet(wallet):

	stdin_ok = True
	desc = 'hexadecimal seed data with checksum'

	def _format(self):
		h = self.seed.data.hex()
		self.ssdata.chksum = make_chksum_6(h)
		self.fmt_data = f'{self.ssdata.chksum} {split_into_cols(4, h)}\n'

	def _deformat(self):
		desc = self.desc
		d = self.fmt_data.split()
		try:
			d[1]
			chk, hstr = d[0], ''.join(d[1:])
		except:
			msg(f'{self.fmt_data.strip()!r}: invalid {desc}')
			return False

		if not len(hstr)*4 in Seed.lens:
			msg(f'Invalid data length ({len(hstr)}) in {desc}')
			return False

		if not is_chksum_6(chk):
			msg(f'{chk!r}: invalid checksum format in {desc}')
			return False

		if not is_hex_str(hstr):
			msg(f'{hstr!r}: not a hexadecimal string, in {desc}')
			return False

		self.cfg._util.vmsg_r(f'Validating {desc} checksum...')

		if not self.cfg._util.compare_chksums(chk, 'file', make_chksum_6(hstr), 'computed', verbose=True):
			return False

		self.seed = Seed(self.cfg, seed_bin=bytes.fromhex(hstr))
		self.ssdata.chksum = chk

		self.check_usr_seed_len()

		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>

"""
test.misc.tool_api_test: test the MMGen suite tool API
"""

import sys
from mmgen.key import PrivKey
from mmgen.addr import CoinAddr

keys = [
	'118089d66b4a5853765e94923abdd5de4616c6e5118089d66b4a5853765e9492',
	'deadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeef',
	]

def check_equal(a, b):
	assert a == b, f'{a} != {b}'

def msg(*args, **kwargs):
	print(*args, **kwargs, file=sys.stdout)

def init_coin(t, coinsym, network, addrtype):
	t.init_coin(coinsym, network)
	t.addrtype = addrtype
	check_equal(type(t.addrtype).__name__, 'MMGenAddrType')
	msg(f'\n{t.coin} {t.proto.network.capitalize()} {t.addrtype.name.upper()} ({t.addrtype})')

def test_randpair(t):
	wif, addr = t.randpair()
	wif_chk = PrivKey(proto=t.proto, wif=wif).wif
	check_equal(wif, wif_chk)
	msg('\n  === randwif ===')
	msg('  wif:', wif)
	msg('  addr:', CoinAddr(proto=t.proto, addr=addr))

def test_wif2addr(t, wif_chk, addr_chk, key_idx):
	key_bytes = bytes.fromhex(keys[key_idx])
	wif = PrivKey(
		proto       = t.proto,
		s           = key_bytes,
		compressed  = t.addrtype.compressed,
		pubkey_type = t.addrtype.pubkey_type).wif
	addr = t.wif2addr(wif)

	msg('\n  === wif2addr ===')
	msg('  wif:', PrivKey(proto=t.proto, wif=wif).wif)
	msg('  addr:', CoinAddr(proto=t.proto, addr=addr))

	addr_ph = t.privhex2addr(key_bytes.hex())

	check_equal(addr, addr_ph)
	check_equal(wif, wif_chk)
	check_equal(addr, addr_chk)

def test_triplet(tool, coin, network, addrtype, key_idx, wif_chk, addr_chk):
	init_coin(tool, coin, network, addrtype)
	test_randpair(tool)
	test_wif2addr(tool, wif_chk, addr_chk, key_idx)

def run_test():

	from mmgen.tool.api import tool_api
	tool = tool_api(cfg)

	tool.coins
	tool.print_addrtypes()

	tool.usr_randchars     # getter
	tool.usr_randchars = 0 # setter
	check_equal(tool.usr_randchars, 0)

	check_equal(f'{tool.coin} {tool.proto.cls_name} {tool.addrtype}', 'BTC mainnet L')

	# test vectors from tooltest2.py:

	test_triplet(tool, 'btc', 'mainnet', 'L', 0,
		'5HwzecKMWD82ppJK3qMKpC7ohXXAwcyAN5VgdJ9PLFaAzpBG4sX',
		'1C5VPtgq9xQ6AcTgMAR3J6GDrs72HC4pS1')

	test_triplet(tool, 'btc', 'mainnet', 'C', 0,
		'KwojSzt1VvW343mQfWQi3J537siAt5ktL2qbuCg1ZyKR8BLQ6UJm',
		'1Kz9fVSUMshzPejpzW9D95kScgA3rY6QxF')

	test_triplet(tool, 'btc', 'mainnet', 'segwit', 0,
		'KwojSzt1VvW343mQfWQi3J537siAt5ktL2qbuCg1ZyKR8BLQ6UJm',
		'3AhjTiWHhVJAi1s5CfKMcLzYps12x3gZhg')

	test_triplet(tool, 'btc', 'mainnet', 'B', 0,
		'KwojSzt1VvW343mQfWQi3J537siAt5ktL2qbuCg1ZyKR8BLQ6UJm',
		'bc1q6pqnfwwakuuejpm9w52ds342f9d5u36v0qnz7c')

	if not 'no_altcoin' in sys.argv:
		test_triplet(tool, 'ltc', 'regtest', 'bech32', 1,
			'cV3ZRqf8PhyfiFwtJfkvGu2qmBsazE1wXoA2A16S3nixb3BTvvVx',
			'rltc1qvmqas4maw7lg9clqu6kqu9zq9cluvllnz4kj9y')

		test_triplet(tool, 'xmr', 'mainnet', 'M', 1,
			'e8164dda6d42bd1e261a3406b2038dcbddadbeefdeadbeefdeadbeefdeadbe0f',
			'41i7saPWA53EoHenmJVRt34dubPxsXwoWMnw8AdMyx4mTD1svf7qYzcVjxxRfteLNdYrAxWUMmiPegFW9EfoNgXx7vDMExv')

		test_triplet(tool, 'etc', 'mainnet', 'E', 1,
			'deadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeef',
			'c96aaa54e2d44c299564da76e1cd3184a2386b8d')

		test_triplet(tool, 'zec', 'mainnet', 'Z', 1,
			'SKxuS56e99jpCeD9mMQ5o63zoGPakNdM9HCvt4Vt2cypvRjCdvGJ',
			'zchFELwBxqsAubsLQ8yZgPCDDGukjXJssgCbiTPwFNmFwn9haLnDatzfhLdZzJT4PcU4o2yr92B52UFirUzEdF6ZYM2gBkM')

from mmgen.cfg import Config

cfg = Config(process_opts=True)

run_test()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.view: Monero wallet ops for the MMGen Suite
"""

from ...color import green
from ...util import msg, ymsg

from ..include import gen_acct_addr_info
from ..rpc import MoneroWalletRPC

from .sync import OpSync

class OpList(OpSync):
	stem = 'sync'

	def gen_body(self, wallets_data):
		for (wallet_fn, wallet_data) in wallets_data.items():
			ad = wallet_data.accts_data['subaddress_accounts']
			yield green(f'Wallet {wallet_fn}:')
			for account in range(len(wallet_data.addrs_data)):
				bal = ad[account]['unlocked_balance']
				if self.cfg.skip_empty_accounts and not bal:
					continue
				yield ''
				yield '  Account #{a} [{b} {c}]'.format(
					a = account,
					b = self.proto.coin_amt(bal, from_unit='atomic').hl(),
					c = self.proto.coin_amt.hlc('XMR'))
				yield from gen_acct_addr_info(self, wallet_data, account, indent='  ')

			yield ''

class OpView(OpSync):
	stem = 'open'
	opts = ()
	wallet_offline = True

	def pre_init_action(self):
		ymsg('Running in offline mode. Balances may be out of date!')

	async def process_wallet(self, d, fn, last):

		self.c.call(
			'open_wallet',
			filename = fn.name,
			password = d.wallet_passwd)

		wallet_height = self.c.call('get_height')['height']
		msg(f'  Wallet height: {wallet_height}')

		self.wallets_data[fn.name] = MoneroWalletRPC(self, d).get_wallet_data(print=False, skip_empty_ok=True)

		if not last:
			self.c.call('close_wallet')

		return True

class OpListview(OpView, OpList):
	pass



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tw.prune: Bitcoin base protocol tracking wallet address list prune class
"""

from ....tw.prune import TwAddressesPrune
from .addresses import BitcoinTwAddresses

class BitcoinTwAddressesPrune(BitcoinTwAddresses, TwAddressesPrune):

	prompt_fs_in = [
		'Sort options: [a]mt, [A]ge, [M]mgen addr, [r]everse',
		'Column options: toggle [D]ays/date/confs/block',
		'Filters: show [E]mpty addrs, [U]sed addrs, all [L]abels',
		'View/Actions: pager [v]iew, [w]ide view, r[e]draw{s}',
		'Pruning: [q]uit pruning, [p]rune, [u]nprune, [c]lear prune list:']
	prompt_fs_repl = {
		'BCH': (1, 'Column options: toggle [D]ays/date/confs/block, cas[h]addr')
	}
	key_mappings = {
		'a':'s_amt',
		'A':'s_age',
		'M':'s_twmmid',
		'r':'s_reverse',
		'D':'d_days',
		'e':'d_redraw',
		'E':'d_showempty',
		'U':'d_showused',
		'L':'d_all_labels',
		'v':'a_view',
		'w':'a_view_detail',
		'p':'a_prune',
		'u':'a_unprune',
		'c':'a_clear_prune_list'}



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.common: Shared Bitcoin functions and constants
"""

import hashlib

b58a = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'

def hash160(in_bytes): # OP_HASH160
	return hashlib.new('ripemd160', hashlib.sha256(in_bytes).digest()).digest()

def hash256(in_bytes): # OP_HASH256
	return hashlib.sha256(hashlib.sha256(in_bytes).digest()).digest()

# From en.bitcoin.it:
#  The Base58 encoding used is home made, and has some differences.
#  Especially, leading zeroes are kept as single zeroes when conversion happens.
# Test: 5JbQQTs3cnoYN9vDYaGY6nhQ1DggVsY4FJNBUfEfpSQqrEp3srk
# The 'zero address':
# 1111111111111111111114oLvT2 (pubkeyhash = '\0'*20)

def b58chk_encode(in_bytes):
	lzeroes = len(in_bytes) - len(in_bytes.lstrip(b'\x00'))
	def do_enc(n):
		while n:
			yield b58a[n % 58]
			n //= 58
	return ('1' * lzeroes) + ''.join(do_enc(int.from_bytes(in_bytes+hash256(in_bytes)[:4], 'big')))[::-1]

def b58chk_decode(s):
	lzeroes = len(s) - len(s.lstrip('1'))
	res = sum(b58a.index(ch) * 58**n for n, ch in enumerate(s[::-1]))
	bl = res.bit_length()
	out = b'\x00' * lzeroes + res.to_bytes(bl//8 + bool(bl%8), 'big')
	if out[-4:] != hash256(out[:-4])[:4]:
		raise ValueError('b58chk_decode(): incorrect checksum')
	return out[:-4]



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_d.xmrwallet: xmrwallet tests for the cmdtest.py test suite
"""

import sys, os, time, re, atexit, asyncio, shutil
from subprocess import run, PIPE
from collections import namedtuple

from mmgen.util import async_run, capfirst, is_int, die, list_gen
from mmgen.obj import MMGenRange
from mmgen.amt import XMRAmt
from mmgen.addrlist import ViewKeyAddrList, KeyAddrList, AddrIdxList

from ..include.common import (
	omsg,
	oqmsg_r,
	ok,
	imsg,
	imsg_r,
	write_data_to_file,
	read_from_file,
	silence,
	end_silence,
	strip_ansi_escapes
)
from .include.common import get_file_with_ext
from .include.proxy import TestProxy
from .base import CmdTestBase

# atexit functions:
def stop_daemons(self):
	for v in self.users.values():
		v.md.stop()

def stop_miner_wallet_daemon(self):
	async_run(self.users['miner'].wd_rpc.stop_daemon())

class CmdTestXMRWallet(CmdTestBase):
	"""
	Monero wallet operations
	"""

	networks = ('xmr',)
	passthru_opts = ()
	tmpdir_nums = [29]
	dfl_random_txs = 3
	color = True
	# Bob’s daemon is stopped via process kill, not RPC, so put Bob last in list:
	#    user     sid      autosign  shift kal_range add_coind_args
	user_data = (
		('miner', '98831F3A', False, 130, '1-2', []),
		('alice', 'FE3C6545', False, 150, '1-4', []),
		('bob',   '1378FC64', False, 140, None,  ['--restricted-rpc']),
	)
	tx_relay_user = 'bob'
	datadir_base = os.path.join('test', 'daemons', 'xmrtest')

	cmd_group = (
		('daemon_version',                'checking daemon version'),
		('gen_kafiles_miner_alice',       'generating key-address files for Miner and Alice'),
		('create_wallets_miner',          'creating Monero wallets (Miner)'),
		('set_label_miner',               'setting an address label (Miner, primary account)'),
		('mine_initial_coins',            'mining initial coins'),
		('create_wallets_alice',          'creating Monero wallets (Alice)'),
		('fund_alice',                    'sending funds'),
		('check_bal_alice',               'mining, checking balance'),

		('sync_wallets_all',              'syncing all wallets'),
		('new_account_alice',             'creating a new account (Alice)'),
		('new_account_alice_label',       'creating a new account (Alice, with label)'),
		('new_address_alice',             'creating a new address (Alice)'),
		('new_address_alice_label',       'creating a new address (Alice, with label)'),
		('remove_label_alice',            'removing an address label (Alice, subaddress)'),
		('set_label_alice',               'setting an address label (Alice, subaddress)'),
		('sync_wallets_selected',         'syncing selected wallets'),

		('sweep_to_wallet',               'sweeping to new account in another wallet'),
		('sweep_to_account',              'sweeping to specific account in same wallet'),
		('sweep_to_wallet_account',       'sweeping to specific account in another wallet'),
		('sweep_to_wallet_account_proxy', 'sweeping to specific account in another wallet (via TX relay + proxy)'),
		('sweep_to_same_account_noproxy', 'sweeping to same account (via TX relay, no proxy)'),
		('transfer_to_miner_proxy',       'transferring funds to Miner (via TX relay + proxy)'),
		('transfer_to_miner_noproxy',     'transferring funds to Miner (via TX relay, no proxy)'),

		('transfer_to_miner_create1',     'transferring funds to Miner (create TX)'),
		('transfer_to_miner_send1',       'transferring funds to Miner (send TX via proxy)'),
		('transfer_to_miner_create2',     'transferring funds to Miner (create TX)'),
		('transfer_to_miner_send2',       'transferring funds to Miner (send TX, no proxy)'),

		('sweep_create_and_send',         'sweeping to new account (create TX + send TX, in stages)'),
		('list_wallets_all',              'listing wallets'),
		('stop_daemons',                  'stopping all wallet and coin daemons'),
	)

	def __init__(self, cfg, trunner, cfgs, spawn):
		CmdTestBase.__init__(self, cfg, trunner, cfgs, spawn)

		if trunner is None:
			return

		from mmgen.protocol import init_proto
		self.proto = init_proto(cfg, 'XMR', network='mainnet')
		self.extra_opts = ['--wallet-rpc-password=passw0rd']
		self.init_users()
		self.init_daemon_args()

		for v in self.users.values():
			run(['mkdir', '-p', v.udir])

		self.tx_relay_daemon_parm = 'localhost:{}'.format(self.users[self.tx_relay_user].md.rpc_port)
		self.tx_relay_daemon_proxy_parm = (
			# must be IP, not 'localhost':
			self.tx_relay_daemon_parm + f':127.0.0.1:{TestProxy.port}')

		if not cfg.no_daemon_stop:
			atexit.register(stop_daemons, self)
			atexit.register(stop_miner_wallet_daemon, self)

		if not cfg.no_daemon_autostart:
			stop_daemons(self)
			time.sleep(0.2)
			if os.path.exists(self.datadir_base):
				shutil.rmtree(self.datadir_base)
			os.makedirs(self.datadir_base)
			TestProxy(self, cfg)
			self.start_daemons()

		self.balance = None

	# init methods

	def init_users(self):
		from mmgen.daemon import CoinDaemon
		from mmgen.proto.xmr.daemon import MoneroWalletDaemon
		from mmgen.proto.xmr.rpc import MoneroRPCClient, MoneroWalletRPCClient
		self.users = {}
		tmpdir_num = self.tmpdir_nums[0]
		ud = namedtuple('user_data', [
			'sid',
			'mmwords',
			'autosign',
			'udir',
			'datadir',
			'kal_range',
			'kafile',
			'walletfile_fs',
			'addrfile_fs',
			'md',
			'md_rpc',
			'wd',
			'wd_rpc',
			'add_coind_args',
		])
		# kal_range must be None, a single digit, or a single hyphenated range
		for (
				user,
				sid,
				autosign,
				shift,
				kal_range,
				add_coind_args) in self.user_data:
			tmpdir = os.path.join('test', 'tmp', str(tmpdir_num))
			udir = os.path.join(tmpdir, user)
			datadir = os.path.join(self.datadir_base, user)
			md = CoinDaemon(
				cfg        = self.cfg,
				proto      = self.proto,
				test_suite = True,
				port_shift = shift,
				opts       = ['online'],
				datadir    = datadir
			)
			md_rpc = MoneroRPCClient(
				cfg    = self.cfg,
				proto  = self.proto,
				host   = 'localhost',
				port   = md.rpc_port,
				user   = None,
				passwd = None,
				test_connection = False,
				daemon = md,
			)
			wd = MoneroWalletDaemon(
				cfg          = self.cfg,
				proto        = self.proto,
				test_suite   = True,
				wallet_dir   = udir,
				user         = 'foo',
				passwd       = 'bar',
				port_shift   = shift,
				monerod_addr = f'127.0.0.1:{md.rpc_port}',
			)
			wd_rpc = MoneroWalletRPCClient(
				cfg             = self.cfg,
				daemon          = wd,
				test_connection = False,
			)
			if autosign:
				kafile_suf = 'vkeys'
				fn_stem    = 'MoneroWatchOnlyWallet'
				kafile_dir = self.asi_online.xmr_dir
			else:
				kafile_suf = 'akeys'
				fn_stem    = 'MoneroWallet'
				kafile_dir = udir
			self.users[user] = ud(
				sid            = sid,
				mmwords        = f'test/ref/{sid}.mmwords',
				autosign       = autosign,
				udir           = udir,
				datadir        = datadir,
				kal_range      = kal_range,
				kafile         = f'{kafile_dir}/{sid}-XMR-M[{kal_range}].{kafile_suf}',
				walletfile_fs  = f'{udir}/{sid}-{{}}-{fn_stem}',
				addrfile_fs    = f'{udir}/{sid}-{{}}-{fn_stem}.address.txt',
				md             = md,
				md_rpc         = md_rpc,
				wd             = wd,
				wd_rpc         = wd_rpc,
				add_coind_args = add_coind_args,
			)

	def init_daemon_args(self):
		common_args = ['--p2p-bind-ip=127.0.0.1', '--fixed-difficulty=1', '--regtest'] # --rpc-ssl-allow-any-cert
		for u in self.users:
			other_ports = [self.users[u2].md.p2p_port for u2 in self.users if u2 != u]
			node_args = [f'--add-exclusive-node=127.0.0.1:{p}' for p in other_ports]
			self.users[u].md.usr_coind_args = (
				common_args
				+ node_args
				+ self.users[u].add_coind_args)

	# cmd_group methods

	def daemon_version(self):
		rpc_port = self.users['miner'].md.rpc_port
		return self.spawn('mmgen-tool', ['--coin=xmr', f'--rpc-port={rpc_port}', 'daemon_version'])

	def gen_kafiles_miner_alice(self):
		return self.gen_kafiles(['miner', 'alice'])

	def gen_kafiles(self, users):
		for user, data in self.users.items():
			if not user in users:
				continue
			run(['mkdir', '-p', data.udir])
			run(f'rm -f {data.kafile}', shell=True)
			t = self.spawn(
				'mmgen-keygen', [
					'-q', '--accept-defaults', '--coin=xmr',
					f'--outdir={data.udir}', data.mmwords, data.kal_range
				],
				extra_desc = f'({capfirst(user)})')
			t.read()
			t.ok()
		t.skip_ok = True
		return t

	def create_wallets_miner(self):
		return self.create_wallets('miner')

	def create_wallets_alice(self):
		return self.create_wallets('alice')

	def create_wallets(self, user, wallet=None, add_opts=[], op='create'):
		assert wallet is None or is_int(wallet), 'wallet arg'
		data = self.users[user]
		stem_glob = data.walletfile_fs.format(wallet or '*')
		for glob in (
				stem_glob,
				stem_glob + '.keys',
				stem_glob + '.address.txt'):
			run(f'rm -f {glob}', shell=True)
		t = self.spawn(
			'mmgen-xmrwallet',
			[f'--wallet-dir={data.udir}']
			+ self.extra_opts
			+ (self.autosign_opts if data.autosign else [])
			+ add_opts
			+ [op]
			+ ([] if data.autosign else [data.kafile])
			+ [wallet or data.kal_range]
		)
		for i in MMGenRange(wallet or data.kal_range).items:
			write_data_to_file(
				self.cfg,
				self.users[user].addrfile_fs.format(i),
				t.expect_getend('Address: '),
				quiet = True
			)
		return t

	def new_addr_alice(self, spec, cfg, expect, kafile=None):
		data = self.users['alice']
		t = self.spawn(
			'mmgen-xmrwallet',
			self.extra_opts
			+ [f'--wallet-dir={data.udir}']
			+ [f'--daemon=localhost:{data.md.rpc_port}']
			+ (['--no-start-wallet-daemon'] if cfg in ('continue', 'stop') else [])
			+ (['--no-stop-wallet-daemon'] if cfg in ('start', 'continue') else [])
			+ ['new', (kafile or data.kafile), spec])
		t.expect(expect, 'y', regex=True)
		return t

	na_idx = 1

	def new_account_alice(self):
		return self.new_addr_alice(
			'4',
			'start',
			r'Creating new account for wallet .*4.* with label .*‘xmrwallet new account .*y/N\): ')

	def new_account_alice_label(self):
		return self.new_addr_alice(
			'4,Alice’s new account',
			'continue',
			r'Creating new account for wallet .*4.* with label .*‘Alice’s new account .*y/N\): ')

	def new_address_alice(self):
		return self.new_addr_alice(
			'4:2',
			'continue',
			r'Creating new address for wallet .*4.*, account .*#2.* with label .*‘xmrwallet new address .*y/N\): ')

	def new_address_alice_label(self):
		return self.new_addr_alice(
			'4:2,Alice’s new address',
			'stop',
			r'Creating new address for wallet .*4.*, account .*#2.* with label .*‘Alice’s new address .*y/N\): ')

	async def mine_initial_coins(self):
		self.spawn(msg_only=True, extra_desc='(opening wallet)')
		await self.open_wallet_user('miner', 1)
		ok()
		# NB: a large balance is required to avoid ‘insufficient outputs’ error
		return await self.mine_chk('miner', 1, 0, lambda x: x.ub > 2000, 'unlocked balance > 2000')

	async def fund_alice(self, wallet=1, amt=1234567891234):
		self.spawn(msg_only=True, extra_desc='(transferring funds from Miner wallet)')
		await self.transfer(
			'miner',
			amt,
			read_from_file(self.users['alice'].addrfile_fs.format(wallet)),
		)
		return 'ok'

	async def check_bal_alice(self, wallet=1, bal='1.234567891234'):
		return await self.mine_chk(
			'alice', wallet, 0,
			lambda x: str(x.ub) == bal, f'unlocked balance == {bal}',
			random_txs = self.dfl_random_txs
		)

	def set_label_miner(self):
		return self.set_label_user('miner', '1:0:0,"Miner’s new primary account label [1:0:0]"', 'updated')

	def remove_label_alice(self):
		return self.set_label_user('alice', '4:2:2,""', 'removed', add_opts=['--full-address'])

	def set_label_alice(self):
		return self.set_label_user('alice', '4:2:2,"Alice’s new subaddress label [4:2:2]"', 'set')

	def set_label_user(self, user, label_spec, expect, add_opts=[]):
		data = self.users[user]
		cmd_opts = [f'--wallet-dir={data.udir}', f'--daemon=localhost:{data.md.rpc_port}']
		t = self.spawn(
			'mmgen-xmrwallet',
			self.extra_opts
			+ add_opts
			+ cmd_opts
			+ ['label', data.kafile, label_spec]
		)
		t.expect('(y/N): ', 'y')
		t.expect(f'Label successfully {expect}')
		return t

	def sync_wallets_all(self):
		return self.sync_wallets('alice', add_opts=['--rescan-blockchain', '-Ee'])

	def sync_wallets_selected(self):
		return self.sync_wallets('alice', wallets='1-2,4', add_opts=['--full-address'])

	def list_wallets_all(self):
		return self.sync_wallets('alice', op='list', add_opts=['-Ee', '--full-address'])

	def sync_wallets_alice(self):
		return self.sync_wallets('alice')

	def sync_wallets_bob(self):
		return self.sync_wallets('bob')

	def sync_wallets_miner(self):
		return self.sync_wallets('miner')

	def sync_wallets(self, user, op='sync', wallets=None, add_opts=[], bal_chk_func=None):
		data = self.users[user]
		if data.autosign:
			self.insert_device_online()
		cmd_opts = list_gen(
			[f'--wallet-dir={data.udir}'],
			[f'--daemon=localhost:{data.md.rpc_port}'],
		)
		t = self.spawn(
			'mmgen-xmrwallet',
			self.extra_opts
			+ cmd_opts
			+ (self.autosign_opts if data.autosign else [])
			+ add_opts
			+ [op]
			+ ([] if data.autosign else [data.kafile])
			+ ([wallets] if wallets else [])
		)
		wlist = AddrIdxList(fmt_str=wallets) if wallets else MMGenRange(data.kal_range).items
		for n, wnum in enumerate(wlist, 1):
			t.expect('ing wallet {}/{} ({})'.format(
				n,
				len(wlist),
				os.path.basename(data.walletfile_fs.format(wnum)),
			))
			if op in ('view', 'listview'):
				t.expect('Wallet height: ')
			else:
				t.expect('Chain height: ')
				t.expect('Wallet height: ')
				res = strip_ansi_escapes(t.expect_getend('Balance: '))
				if bal_chk_func:
					m = re.match(r'(\S+) Unlocked balance: (\S+)', res, re.DOTALL)
					amts = [XMRAmt(amt) for amt in m.groups()]
					assert bal_chk_func(n, *amts), f'balance check for wallet {n} failed!'
		if data.autosign:
			t.read()
			self.remove_device_online()
		return t

	def do_op(
			self,
			op,
			user,
			arg2,
			tx_relay_parm = None,
			no_relay      = False,
			use_existing  = False,
			add_opts      = [],
			add_desc      = None,
			do_ret        = False):

		data = self.users[user]
		cmd_opts = list_gen(
			[f'--wallet-dir={data.udir}'],
			[f'--outdir={data.udir}', not data.autosign],
			[f'--daemon=localhost:{data.md.rpc_port}'],
			[f'--tx-relay-daemon={tx_relay_parm}', tx_relay_parm],
			['--no-relay', no_relay and not data.autosign],
		)
		add_desc = (', ' + add_desc) if add_desc else ''

		t = self.spawn(
			'mmgen-xmrwallet',
			self.extra_opts
			+ cmd_opts
			+ add_opts
			+ (self.autosign_opts if data.autosign else [])
			+ [op]
			+ ([] if data.autosign else [data.kafile])
			+ [arg2],
			extra_desc = f'({capfirst(user)}{add_desc})')

		if op == 'sign':
			return t

		if op in ('sweep', 'sweep_all'):
			desc = 'address' if re.match(r'.*:\d+$', arg2) else 'account'
			t.expect(rf'Create new {desc} .* \(y/N\): ', ('y', 'n')[use_existing], regex=True)
			if use_existing:
				t.expect(rf'to last existing {desc} .* \(y/N\): ', 'y', regex=True)

		dtype = 'unsigned' if data.autosign else 'signed'
		t.expect(f'Save {dtype} transaction? (y/N): ', 'y')
		t.written_to_file(f'{dtype.capitalize()} transaction')

		if not no_relay:
			t.expect(f'Relay {op} transaction? (y/N): ', 'y')
			get_file_with_ext(self.users[user].udir, 'sigtx', delete_all=True)

		t.read()

		return t if do_ret else t.ok()

	def sweep_to_wallet(self):
		self.do_op('sweep', 'alice', '1:0,2')
		return self.mine_chk('alice', 2, 1, lambda x: x.ub > 1, 'unlocked balance > 1')

	def sweep_to_account(self):
		self.do_op('sweep', 'alice', '2:1,2:0', use_existing=True)
		return self.mine_chk('alice', 2, 0, lambda x: x.ub > 1, 'unlocked balance > 1')

	def sweep_to_wallet_account(self):
		self.do_op('sweep', 'alice', '2:0,3:0', use_existing=True, add_opts=['-Ee', '--full-address'])
		return self.mine_chk('alice', 3, 0, lambda x: x.ub > 1, 'unlocked balance > 1')

	def sweep_to_wallet_account_proxy(self):
		self.do_op('sweep', 'alice', '3:0,2:1', self.tx_relay_daemon_proxy_parm, add_opts=['--priority=3', '-Ee'])
		return self.mine_chk('alice', 2, 1, lambda x: x.ub > 1, 'unlocked balance > 1')

	def sweep_to_same_account_noproxy(self):
		self.do_op('sweep', 'alice', '2:1', self.tx_relay_daemon_parm)
		return self.mine_chk('alice', 2, 1, lambda x: x.ub > 0.9, 'unlocked balance > 0.9')

	async def transfer_to_miner_proxy(self):
		addr = read_from_file(self.users['miner'].addrfile_fs.format(2))
		amt = '0.135'
		self.do_op('transfer', 'alice', f'2:1:{addr},{amt}', self.tx_relay_daemon_proxy_parm)
		await self.stop_wallet_user('miner')
		await self.open_wallet_user('miner', 2)
		await self.mine_chk('miner', 2, 0, lambda x: str(x.ub) == amt, f'unlocked balance == {amt}')
		ok()
		return await self.mine_chk('alice', 2, 1, lambda x: x.ub > 0.9, 'unlocked balance > 0.9')

	async def transfer_to_miner_noproxy(self):
		addr = read_from_file(self.users['miner'].addrfile_fs.format(2))
		self.do_op('transfer', 'alice', f'2:1:{addr},0.0995', self.tx_relay_daemon_parm, add_opts=['--full-address'])
		await self.mine_chk('miner', 2, 0, lambda x: str(x.ub) == '0.2345', 'unlocked balance == 0.2345')
		ok()
		return await self.mine_chk('alice', 2, 1, lambda x: x.ub > 0.9, 'unlocked balance > 0.9')

	def transfer_to_miner_create(self, amt):
		get_file_with_ext(self.users['alice'].udir, 'sigtx', delete_all=True)
		addr = read_from_file(self.users['miner'].addrfile_fs.format(2))
		return self.do_op('transfer', 'alice', f'2:1:{addr},{amt}', no_relay=True, do_ret=True, add_opts=['-Ee'])

	def transfer_to_miner_create1(self):
		return self.transfer_to_miner_create('0.0111')

	def transfer_to_miner_create2(self):
		return self.transfer_to_miner_create('0.0012')

	def relay_tx(self, relay_opt, add_desc=None):
		user = 'alice'
		data = self.users[user]
		add_desc = (', ' + add_desc) if add_desc else ''
		t = self.spawn(
			'mmgen-xmrwallet',
			self.extra_opts
			+ [relay_opt, 'relay', get_file_with_ext(data.udir, 'sigtx')],
			extra_desc = f'(relaying TX, {capfirst(user)}{add_desc})')
		t.expect('Relay transaction? ', 'y')
		t.read()
		t.ok()
		return t

	async def transfer_to_miner_send1(self):
		self.relay_tx(f'--tx-relay-daemon={self.tx_relay_daemon_proxy_parm}', add_desc='via proxy')
		await self.mine_chk('miner', 2, 0, lambda x: str(x.ub) == '0.2456', 'unlocked balance == 0.2456')
		ok()
		return await self.mine_chk('alice', 2, 1, lambda x: x.ub > 0.9, 'unlocked balance > 0.9')

	async def transfer_to_miner_send2(self):
		self.relay_tx(f'--tx-relay-daemon={self.tx_relay_daemon_parm}', add_desc='no proxy')
		await self.mine_chk('miner', 2, 0, lambda x: str(x.ub) == '0.2468', 'unlocked balance == 0.2468')
		ok()
		return await self.mine_chk('alice', 2, 1, lambda x: x.ub > 0.9, 'unlocked balance > 0.9')

	async def sweep_create_and_send(self):
		get_file_with_ext(self.users['alice'].udir, 'sigtx', delete_all=True)

		self.do_op('sweep_all', 'alice', '2:1,3', no_relay=True, use_existing=True)
		ok()

		self.relay_tx(f'--tx-relay-daemon={self.tx_relay_daemon_parm}')

		min_bal = XMRAmt('0.9')
		return await self.mine_chk('alice', 3, 0, lambda x: x.ub > min_bal, f'bal > {min_bal}')

	# wallet methods

	async def open_wallet_user(self, user, wnum):
		data = self.users[user]
		if data.autosign:
			self.insert_device_online()
			self.do_mount_online()
		silence()
		kal = (ViewKeyAddrList if data.autosign else KeyAddrList)(
			cfg      = self.cfg,
			proto    = self.proto,
			infile   = data.kafile,
			skip_chksum_msg = True,
			key_address_validity_check = False)
		end_silence()
		if data.autosign:
			self.do_umount_online()
			self.remove_device_online()
		self.users[user].wd.start(silent=self.tr.quiet)
		return data.wd_rpc.call(
			'open_wallet',
			filename = os.path.basename(data.walletfile_fs.format(wnum)),
			password = kal.entry(wnum).wallet_passwd)

	async def stop_wallet_user(self, user):
		await self.users[user].wd_rpc.stop_daemon(silent=self.tr.quiet)
		return 'ok'

	# mining methods

	async def mine5(self):
		return await self.mine(5)

	async def _get_height(self):
		u = self.users['miner']
		for _ in range(20):
			try:
				return u.md_rpc.call('get_last_block_header')['block_header']['height']
			except Exception as e:
				if 'onnection refused' in str(e):
					omsg(f'{e}\nMonerod appears to have crashed. Attempting to restart...')
					await asyncio.sleep(5)
					u.md.restart()
					await asyncio.sleep(5)
					await self.start_mining()
				else:
					raise
		die(2, 'Restart attempt limit exceeded')

	async def mine10(self):
		return await self.mine(10)

	async def mine30(self):
		return await self.mine(30)

	async def mine100(self):
		return await self.mine(100)

	async def mine(self, nblks):
		start_height = height = await self._get_height()
		imsg(f'Height: {height}')
		imsg_r(f'Mining {nblks} blocks...')
		await self.start_mining()
		while height < start_height + nblks:
			await asyncio.sleep(2)
			height = await self._get_height()
			imsg_r('.')
		ret = await self.stop_mining()
		imsg('done')
		imsg(f'Height: {height}')
		return 'ok' if ret == 'OK' else False

	async def start_mining(self):
		data = self.users['miner']
		addr = read_from_file(data.addrfile_fs.format(1)) # mine to wallet #1, account 0

		for _ in range(20):
			# NB: threads_count > 1 provides no benefit and leads to connection errors with MSWin/MSYS2
			ret = data.md_rpc.call_raw(
				'start_mining',
				do_background_mining = False, # run mining in background or foreground
				ignore_battery       = True,  # ignore battery state (on laptop)
				miner_address        = addr,  # account address to mine to
				threads_count        = 1)    # number of mining threads to run
			status = self.get_status(ret)
			if status == 'OK':
				return True
			elif status == 'BUSY':
				await asyncio.sleep(5)
				omsg('Daemon busy.  Attempting to start mining...')
			else:
				die(2, f'Monerod returned status {status}')
		die(2, 'Max retries exceeded')

	async def stop_mining(self):
		ret = self.users['miner'].md_rpc.call_raw('stop_mining')
		return self.get_status(ret)

	async def mine_chk(
			self,
			user,
			wnum,
			account,
			test,
			test_desc,
			test2      = None,
			test2_desc = None,
			random_txs = None,
			return_bal = False):

		"""
		- open destination wallet
		- optionally create and broadcast random TXs
		- start mining
		- mine until funds appear in wallet
		- stop mining
		- close wallet
		"""

		async def send_random_txs():
			from mmgen.tool.api import tool_api
			t = tool_api(self.cfg)
			t.init_coin('XMR', 'mainnet')
			t.usr_randchars = 0
			imsg_r('Sending random transactions: ')
			for i in range(random_txs):
				await self.transfer(
					'miner',
					123456789,
					t.randpair()[1],
				)
				imsg_r(f'{i+1} ')
				oqmsg_r('+')
				await asyncio.sleep(0.5)
			imsg('')

		def print_balance(dest, bal_info):
			imsg('Total balances in {}’s wallet {}, account #{}: {} (total), {} (unlocked)'.format(
				capfirst(dest.user),
				dest.wnum,
				dest.account,
				bal_info.b.hl(),
				bal_info.ub.hl(),
			))

		async def get_balance(dest, count):
			data = self.users[dest.user]
			data.wd_rpc.call('refresh')
			if count and not count % 20:
				data.wd_rpc.call('rescan_blockchain')
			ret = data.wd_rpc.call('get_accounts')['subaddress_accounts'][dest.account]
			d_tup = namedtuple('bal_info', ['b', 'ub'])
			return d_tup(
				b  = XMRAmt(ret['balance'], from_unit='atomic'),
				ub = XMRAmt(ret['unlocked_balance'], from_unit='atomic'))

		# start execution:

		self.do_msg(extra_desc =
			(f'sending {random_txs} random TXs, ' if random_txs else '') +
			f'mining, checking wallet {user}:{wnum}:{account}')

		dest = namedtuple(
			'dest_info', ['user', 'wnum', 'account', 'test', 'test_desc', 'test2', 'test2_desc'])(
				user, wnum, account, test, test_desc, test2, test2_desc)

		if dest.user != 'miner':
			await self.open_wallet_user(dest.user, dest.wnum)

		bal_info_start = await get_balance(dest, 0)
		chk_bal_chg = dest.test(bal_info_start) == 'chk_bal_chg'

		if random_txs:
			await send_random_txs()

		await self.start_mining()

		h = await self._get_height()
		imsg_r(f'Chain height: {h} ')

		max_iterations, min_height = (300, 64) if sys.platform == 'win32' else (50, 300)
		verbose = False

		for count in range(max_iterations):
			bal_info = await get_balance(dest, count)
			if h > min_height:
				if dest.test(bal_info) is True or (chk_bal_chg and bal_info.ub != bal_info_start.ub):
					imsg('')
					oqmsg_r('+')
					print_balance(dest, bal_info)
					if dest.test2:
						assert dest.test2(bal_info) is True, f'test failed: {dest.test2_desc} ({bal_info})'
					break
			await asyncio.sleep(2)
			h = await self._get_height()
			if count > 12: # something might have gone wrong, so be more verbose
				if not verbose:
					imsg('')
				imsg_r(f'Height: {h}, ')
				print_balance(dest, bal_info)
				verbose = True
			else:
				imsg_r(f'{h} ')
				oqmsg_r('+')
		else:
			die(2, f'Timeout exceeded, balance {bal_info.ub!r}')

		await self.stop_mining()

		if user != 'miner':
			await self.stop_wallet_user(dest.user)

		return bal_info if return_bal else 'ok'

	# util methods

	def get_status(self, ret):
		if ret['status'] != 'OK':
			imsg('RPC status: {}'.format(ret['status']))
		return ret['status']

	def do_msg(self, extra_desc=None):
		self.spawn(msg_only=True, extra_desc=f'({extra_desc})' if extra_desc else None)

	async def transfer(self, user, amt, addr):
		return self.users[user].wd_rpc.call('transfer', destinations=[{'amount':amt, 'address':addr}])

	# daemon start/stop methods

	def start_daemons(self):
		for v in self.users.values():
			run(['mkdir', '-p', v.datadir])
			v.md.start()

	def stop_daemons(self):
		self.spawn(msg_only=True)
		if self.cfg.no_daemon_stop:
			omsg('[not stopping daemons at user request]')
		else:
			omsg('')
			stop_daemons(self)
			atexit.unregister(stop_daemons)

			stop_miner_wallet_daemon(self)
			atexit.unregister(stop_miner_wallet_daemon)
		return 'silent'



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.secp256k1.keygen: secp256k1 public key generation backends for the MMGen suite
"""

from ...key import PubKey
from ...keygen import keygen_base

def pubkey_format(vk_bytes, compressed):
	# if compressed, discard Y coord, replace with appropriate version byte
	# even y: <0, odd y: >0 -- https://bitcointalk.org/index.php?topic=129652.0
	return (b'\x02', b'\x03')[vk_bytes[-1] & 1] + vk_bytes[:32] if compressed else b'\x04' + vk_bytes

class backend:

	class libsecp256k1(keygen_base):

		production_safe = True

		def __init__(self, cfg):
			super().__init__(cfg)
			from .secp256k1 import pubkey_gen
			self.pubkey_gen = pubkey_gen

		def to_pubkey(self, privkey):
			return PubKey(
				s = self.pubkey_gen(privkey, int(privkey.compressed)),
				compressed = privkey.compressed)

		@classmethod
		def get_clsname(cls, cfg, *, silent=False):
			try:
				from .secp256k1 import pubkey_gen
				if not pubkey_gen(bytes.fromhex('deadbeef'*8), 1):
					from ...util import die
					die('ExtensionModuleError',
						'Unable to execute pubkey_gen() from secp256k1 extension module')
				return cls.__name__
			except ImportError as e:
				if not silent:
					from ...util import ymsg
					ymsg(str(e))
				cfg._util.qmsg('Using (slow) native Python ECDSA library for public key generation')
				return 'python_ecdsa'

	class python_ecdsa(keygen_base):

		production_safe = False

		def __init__(self, cfg):
			super().__init__(cfg)
			import ecdsa
			self.ecdsa = ecdsa

		def to_pubkey(self, privkey):
			"""
			devdoc/guide_wallets.md:
			Uncompressed public keys start with 0x04; compressed public keys begin with 0x03 or
			0x02 depending on whether they're greater or less than the midpoint of the curve.
			"""
			def privnum2pubkey(numpriv, *, compressed=False):
				pk = self.ecdsa.SigningKey.from_secret_exponent(numpriv, curve=self.ecdsa.SECP256k1)
				# vk_bytes = x (32 bytes) + y (32 bytes) (unsigned big-endian)
				return pubkey_format(pk.verifying_key.to_string(), compressed)

			return PubKey(
				s = privnum2pubkey(int.from_bytes(privkey, 'big'), compressed=privkey.compressed),
				compressed = privkey.compressed)



#!/usr/bin/env python3

import sys, os, atexit
pn = os.path.abspath(os.path.dirname(sys.argv[0]))
parpar = os.path.dirname(os.path.dirname(pn))
os.chdir(parpar)
sys.path[0] = os.curdir

from mmgen.cfg import Config
from mmgen.util import msg
from mmgen.ui import keypress_confirm
from mmgen.led import LEDControl

opts_data = {
	'text': {
		'desc': 'Interactively test LED functionality',
		'usage': 'command',
		'options': """
-h, --help     Print this help message
""",
	}
}

cfg = Config(opts_data=opts_data)

def confirm_or_exit(prompt):
	keypress_confirm(cfg, f'{prompt}.  OK?', default_yes=True, do_exit=True)

confirm_or_exit('This script will interactively test LED functionality')

led = LEDControl(enabled=True)

color = led.board.color.capitalize()

atexit.register(led.stop)

confirm_or_exit(f'{color} LED should now be turned off')

led.set('busy')

confirm_or_exit(f'{color} LED should now be signaling busy (rapid flashing)')

led.set('standby')

confirm_or_exit(f'{color} LED should now be signaling standby (slow flashing)')

led.set('error')

confirm_or_exit(f'{color} LED should now be signaling error (insistent flashing)')

led.set('off')

confirm_or_exit(f'{color} LED should now be turned off')

led.stop()

confirm_or_exit(f'{color} LED should now be in its original state [trigger={led.board.trigger_reset}]')



import os as overlay_fake_os
from . import view_orig as overlay_fake_orig_mod

from .view_orig import *

overlay_fake_orig_mod.CUR_HOME  = '\n[CUR_HOME]\n'
overlay_fake_orig_mod.CUR_UP    = lambda n: f'\n[CUR_UP({n})]\n'
overlay_fake_orig_mod.CUR_DOWN  = lambda n: f'\n[CUR_DOWN({n})]\n'
overlay_fake_orig_mod.ERASE_ALL = '\n[ERASE_ALL]\n'

if overlay_fake_os.getenv('MMGEN_TEST_SUITE_DETERMINISTIC'):

	class overlay_fake_data:
		# add a minute to each successive time value
		time_iter = (1862651471 + (i*60) for i in range(1000000))

	TwView.date_formatter = {
		'days':      lambda rpc, secs: (next(overlay_fake_data.time_iter) - secs) // 86400,
		'date':      lambda rpc, secs: '{}-{:02}-{:02}'.format(
			*time.gmtime(next(overlay_fake_data.time_iter))[:3])[2:],
		'date_time': lambda rpc, secs: '{}-{:02}-{:02} {:02}:{:02}'.format(
			*time.gmtime(next(overlay_fake_data.time_iter))[:5]),
	}

if overlay_fake_os.getenv('MMGEN_BOGUS_UNSPENT_DATA'):

	# 1831006505 (09 Jan 2028) = projected time of block 1000000
	TwView.date_formatter['days'] = lambda rpc, secs: (2178144000 - secs) // 86400 # 9 Jan 2039 00:00:00



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
ui: Interactive user interface functions for the MMGen suite
"""

import sys, os

from .util import msg, msg_r, Msg, die

def confirm_or_raise(cfg, message, action, *, expect='YES', exit_msg='Exiting at user request'):
	if message:
		msg(message)
	if line_input(
			cfg,
			(f'{action}  ' if action[0].isupper() else f'Are you sure you want to {action}?\n') +
			f'Type uppercase {expect!r} to confirm: '
		).strip() != expect:
		die('UserNonConfirmation', exit_msg)

def get_words_from_user(cfg, prompt):
	words = line_input(cfg, prompt, echo=cfg.echo_passphrase).split()
	if cfg.debug:
		msg('Sanitized input: [{}]'.format(' '.join(words)))
	return words

def get_data_from_user(cfg, *, desc='data'): # user input MUST be UTF-8
	data = line_input(cfg, f'Enter {desc}: ', echo=cfg.echo_passphrase)
	if cfg.debug:
		msg(f'User input: [{data}]')
	return data

def line_input(cfg, prompt, *, echo=True, insert_txt='', hold_protect=True):
	"""
	multi-line prompts OK
	one-line prompts must begin at beginning of line
	empty prompts forbidden due to interactions with readline
	"""
	assert prompt, 'calling line_input() with an empty prompt forbidden'

	def get_readline():
		try:
			import readline
			return readline
		except ImportError:
			return False

	if hold_protect:
		from .term import kb_hold_protect
		kb_hold_protect()

	if cfg.test_suite_popen_spawn:
		msg(prompt)
		sys.stderr.flush() # required by older Pythons (e.g. v3.7)
		reply = os.read(0, 4096).decode().rstrip('\n') # strip NL to mimic behavior of input()
	elif not sys.stdin.isatty():
		msg_r(prompt)
		reply = input('')
	elif echo:
		readline = get_readline()
		if readline and insert_txt:
			readline.set_startup_hook(lambda: readline.insert_text(insert_txt))
		reply = input(prompt)
		if readline and insert_txt:
			readline.set_startup_hook(lambda: readline.insert_text(''))
	else:
		from getpass import getpass
		reply = getpass(prompt)

	if hold_protect:
		kb_hold_protect()

	return reply.strip()

def keypress_confirm(
		cfg,
		prompt,
		*,
		default_yes     = False,
		verbose         = False,
		no_nl           = False,
		complete_prompt = False,
		do_exit         = False,
		exit_msg        = 'Exiting at user request'):

	def do_return(retval):
		if do_exit and not retval:
			die(1, exit_msg)
		return retval

	if not complete_prompt:
		prompt = '{} {}: '.format(prompt, '(Y/n)' if default_yes else '(y/N)')

	nl = f'\r{" "*len(prompt)}\r' if no_nl else '\n'

	if cfg.accept_defaults:
		msg(prompt)
		return do_return(default_yes)

	from .term import get_char
	while True:
		reply = get_char(prompt, immed_chars='yYnN').strip('\n\r')
		if not reply:
			msg_r(nl)
			return do_return(default_yes)
		elif reply in 'yYnN':
			msg_r(nl)
			return do_return(reply in 'yY')
		else:
			msg_r('\nInvalid reply\n' if verbose else '\r')

def do_pager(text):

	pagers = ['less', 'more']
	end_msg = '\n(end of text)\n\n'
	os.environ['LESS'] = '--jump-target=2 --shift=4 --tabs=4 --RAW-CONTROL-CHARS --chop-long-lines'

	if 'PAGER' in os.environ and os.environ['PAGER'] != pagers[0]:
		pagers = [os.environ['PAGER']] + pagers

	from subprocess import run
	from .color import set_vt100
	for pager in pagers:
		try:
			m = text + ('' if pager == 'less' else end_msg)
			run([pager], input=m.encode(), check=True)
			msg_r('\r')
		except:
			pass
		else:
			break
	else:
		Msg(text+end_msg)
	set_vt100()

def do_license_msg(cfg, *, immed=False):

	if cfg.quiet or cfg.no_license or cfg.yes or not cfg.stdin_tty:
		return

	from .contrib import license as gpl
	from .cfg import gc
	msg(gpl.warning.format(gc=gc))

	from .term import get_char
	prompt = "Press 'w' for conditions and warranty info, or 'c' to continue: "
	while True:
		reply = get_char(prompt, immed_chars=('', 'wc')[bool(immed)])
		if reply == 'w':
			do_pager(gpl.conditions)
		elif reply == 'c':
			msg('')
			break
		else:
			msg_r('\r')
	msg('')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
autosign: Auto-sign MMGen transactions, message files and XMR wallet output files
"""

import sys, os, asyncio
from stat import S_IWUSR, S_IRUSR
from pathlib import Path
from subprocess import run, PIPE, DEVNULL

from .cfg import Config
from .util import msg, msg_r, ymsg, rmsg, gmsg, bmsg, die, suf, fmt, fmt_list, is_int, have_sudo, capfirst
from .color import yellow, red, orange, brown, blue
from .wallet import Wallet, get_wallet_cls
from .addrlist import AddrIdxList
from .filename import find_file_in_dir
from .fileutil import shred_file
from .ui import keypress_confirm

def SwapMgr(*args, **kwargs):
	if sys.platform == 'linux':
		return SwapMgrLinux(*args, **kwargs)
	elif sys.platform == 'darwin':
		return SwapMgrMacOS(*args, **kwargs)

class SwapMgrBase:

	def __init__(self, cfg, *, ignore_zram=False):
		self.cfg = cfg
		self.ignore_zram = ignore_zram
		self.desc = 'disk swap' if ignore_zram else 'swap'

	def enable(self, *, quiet=False):
		ret = self.do_enable()
		if not quiet:
			self.cfg._util.qmsg(
				f'{capfirst(self.desc)} successfully enabled' if ret else
				f'{capfirst(self.desc)} is already enabled' if ret is None else
				f'Could not enable {self.desc}')
		return ret

	def disable(self, *, quiet=False):
		self.cfg._util.qmsg_r(f'Attempting to disable {self.desc}...')
		ret = self.do_disable()
		self.cfg._util.qmsg('success')
		if not quiet:
			self.cfg._util.qmsg(
				f'{capfirst(self.desc)} successfully disabled ({fmt_list(ret, fmt="no_quotes")})'
					if ret and isinstance(ret, list) else
				f'{capfirst(self.desc)} successfully disabled' if ret else
				f'No active {self.desc}')
		return ret

	def process_cmds(self, op, cmds):
		if not cmds:
			return
		if have_sudo(silent=True) and not self.cfg.test_suite:
			for cmd in cmds:
				run(cmd.split(), check=True)
		else:
			pre = 'failure\n' if op == 'disable' else ''
			fs = blue('{a} {b} manually by executing the following command{c}:\n{d}')
			post = orange('[To prevent this message in the future, enable sudo without a password]')
			m = pre + fs.format(
				a = 'Please disable' if op == 'disable' else 'Enable',
				b = self.desc,
				c = suf(cmds),
				d = fmt_list(cmds, indent='  ', fmt='col')) + '\n' + post
			msg(m)
			if not self.cfg.test_suite:
				sys.exit(1)

class SwapMgrLinux(SwapMgrBase):

	def get_active(self):
		for cmd in ('/sbin/swapon', 'swapon'):
			try:
				cp = run([cmd, '--show=NAME', '--noheadings'], stdout=PIPE, text=True, check=True)
				break
			except Exception:
				if cmd == 'swapon':
					raise
		res = cp.stdout.splitlines()
		return [e for e in res if not e.startswith('/dev/zram')] if self.ignore_zram else res

	def do_enable(self):
		if ret := self.get_active():
			ymsg(f'Warning: {self.desc} is already enabled: ({fmt_list(ret, fmt="no_quotes")})')
		self.process_cmds('enable', ['sudo swapon --all'])
		return True

	def do_disable(self):
		swapdevs = self.get_active()
		if not swapdevs:
			return None
		self.process_cmds('disable', [f'sudo swapoff {swapdev}' for swapdev in swapdevs])
		return swapdevs

class SwapMgrMacOS(SwapMgrBase):

	def get_active(self):
		cmd = 'launchctl print system/com.apple.dynamic_pager'
		return run(cmd.split(), stdout=DEVNULL, stderr=DEVNULL).returncode == 0

	def _do_action(self, active, op, cmd):
		if self.get_active() is active:
			return None
		else:
			cmd = f'sudo launchctl {cmd} -w /System/Library/LaunchDaemons/com.apple.dynamic_pager.plist'
			self.process_cmds(op, [cmd])
			return True

	def do_enable(self):
		return self._do_action(active=True, op='enable', cmd='load')

	def do_disable(self):
		return self._do_action(active=False, op='disable', cmd='unload')

class Signable:

	non_xmr_signables = (
		'transaction',
		'automount_transaction',
		'message')

	xmr_signables = (              # order is important!
		'xmr_wallet_outputs_file', # import XMR wallet outputs BEFORE signing transactions
		'xmr_transaction')

	class base:

		clean_all = False
		multiple_ok = True
		action_desc = 'signed'

		def __init__(self, parent):
			self.parent = parent
			self.cfg = parent.cfg
			self.dir = getattr(parent, self.dir_name)
			self.name = type(self).__name__

		@property
		def submitted(self):
			return self._processed('_submitted', self.subext)

		def _processed(self, attrname, ext):
			if not hasattr(self, attrname):
				setattr(self, attrname, tuple(f for f in sorted(self.dir.iterdir()) if f.name.endswith('.'+ext)))
			return getattr(self, attrname)

		@property
		def unsigned(self):
			return self._unprocessed('_unsigned', self.rawext, self.sigext)

		@property
		def unsubmitted(self):
			return self._unprocessed('_unsubmitted', self.sigext, self.subext)

		@property
		def unsubmitted_raw(self):
			return self._unprocessed('_unsubmitted_raw', self.rawext, self.subext)

		unsent = unsubmitted
		unsent_raw = unsubmitted_raw

		def _unprocessed(self, attrname, rawext, sigext):
			if not hasattr(self, attrname):
				dirlist = sorted(self.dir.iterdir())
				names = {f.name for f in dirlist}
				setattr(
					self,
					attrname,
					tuple(f for f in dirlist
						if f.name.endswith('.' + rawext)
							and f.name[:-len(rawext)] + sigext not in names))
			return getattr(self, attrname)

		def print_bad_list(self, bad_files):
			msg('\n{a}\n{b}'.format(
				a = red(f'Failed {self.desc}s:'),
				b = '  {}\n'.format('\n  '.join(self.gen_bad_list(sorted(bad_files, key=lambda f: f.name))))
			))

		def die_wrong_num_txs(self, tx_type, *, msg=None, desc=None, show_dir=False):
			num_txs = len(getattr(self, tx_type))
			die('AutosignTXError', "{m}{a} {b} transaction{c} {d} {e}!".format(
				m = msg + '\n' if msg else '',
				a = 'One' if num_txs == 1 else 'More than one' if num_txs else 'No',
				b = desc or tx_type,
				c = suf(num_txs),
				d = 'already present' if num_txs else 'present',
				e = f'in ‘{getattr(self.parent, self.dir_name)}’' if show_dir else 'on removable device',
			))

		def check_create_ok(self):
			if len(self.unsigned):
				self.die_wrong_num_txs('unsigned', msg='Cannot create transaction')
			if len(self.unsent):
				die('AutosignTXError', 'Cannot create transaction: you have an unsent transaction')

		def get_unsubmitted(self, tx_type='unsubmitted'):
			if len(self.unsubmitted) == 1:
				return self.unsubmitted[0]
			else:
				self.die_wrong_num_txs(tx_type)

		def get_unsent(self):
			return self.get_unsubmitted('unsent')

		def get_submitted(self):
			if len(self.submitted) == 0:
				self.die_wrong_num_txs('submitted')
			else:
				return self.submitted

		def get_abortable(self):
			if len(self.unsent_raw) != 1:
				self.die_wrong_num_txs('unsent_raw', desc='unsent')
			if len(self.unsent) > 1:
				self.die_wrong_num_txs('unsent')
			if self.unsent:
				if self.unsent[0].stem != self.unsent_raw[0].stem:
					die(1, f'{self.unsent[0]}, {self.unsent_raw[0]}: file mismatch')
			return self.unsent_raw + self.unsent

		def shred_abortable(self):
			files = self.get_abortable() # raises AutosignTXError if no unsent TXs available
			keypress_confirm(
				self.cfg,
				'The following file{} will be securely deleted:\n{}\nOK?'.format(
					suf(files),
					fmt_list(map(str, files), fmt='col', indent='  ')),
					do_exit = True)
			for fn in files:
				msg(f'Shredding file ‘{fn}’')
				shred_file(self.cfg, fn, iterations=15)
			sys.exit(0)

		async def get_last_created(self):
			from .tx import CompletedTX
			ext = '.' + Signable.automount_transaction.subext
			files = [f for f in self.dir.iterdir() if f.name.endswith(ext)]
			return sorted(
				[await CompletedTX(cfg=self.cfg, filename=str(txfile), quiet_open=True) for txfile in files],
				key = lambda x: x.timestamp)[-1]

	class transaction(base):
		desc = 'non-automount transaction'
		rawext = 'rawtx'
		sigext = 'sigtx'
		dir_name = 'tx_dir'
		fail_msg = 'failed to sign'

		async def sign(self, f):
			from .tx import UnsignedTX
			tx1 = UnsignedTX(
					cfg       = self.cfg,
					filename  = f,
					automount = self.name=='automount_transaction')
			if tx1.proto.sign_mode == 'daemon':
				from .rpc import rpc_init
				tx1.rpc = await rpc_init(self.cfg, tx1.proto, ignore_wallet=True)
			from .tx.keys import TxKeys
			tx2 = await tx1.sign(
				TxKeys(
					self.cfg,
					tx1,
					seedfiles = self.parent.wallet_files[:],
					keylist = self.parent.keylist,
					passwdfile = str(self.parent.keyfile),
					autosign = True).keys)
			if tx2:
				tx2.file.write(ask_write=False, outdir=self.dir)
				return tx2
			else:
				return False

		def print_summary(self, signables):

			if self.cfg.full_summary:
				bmsg('\nAutosign summary:\n')
				msg_r('\n'.join(tx.info.format(terse=True) for tx in signables))
				return

			def gen():
				for tx in signables:
					non_mmgen = [o for o in tx.outputs if not o.mmid]
					if non_mmgen:
						yield (tx, non_mmgen)

			body = list(gen())

			if body:
				bmsg('\nAutosign summary:')
				fs = '{}  {} {}'
				t_wid, a_wid = 6, 44

				def gen():
					yield fs.format('TX ID ', 'Non-MMGen outputs'+' '*(a_wid-17), 'Amount')
					yield fs.format('-'*t_wid, '-'*a_wid, '-'*7)
					for tx, non_mmgen in body:
						for nm in non_mmgen:
							yield fs.format(
								tx.txid.fmt(t_wid, color=True) if nm is non_mmgen[0] else ' '*t_wid,
								nm.addr.fmt(nm.addr.view_pref, a_wid, color=True),
								nm.amt.hl() + ' ' + yellow(tx.coin))

				msg('\n' + '\n'.join(gen()))
			else:
				msg('\nNo non-MMGen outputs')

		def gen_bad_list(self, bad_files):
			for f in bad_files:
				yield red(f.name)

	class automount_transaction(transaction):
		desc   = 'automount transaction'
		dir_name = 'txauto_dir'
		rawext = 'arawtx'
		sigext = 'asigtx'
		subext = 'asubtx'
		multiple_ok = False

	class xmr_signable(transaction): # mixin class

		def need_daemon_restart(self, m, new_idx):
			old_idx = self.parent.xmr_cur_wallet_idx
			self.parent.xmr_cur_wallet_idx = new_idx
			return old_idx != new_idx or m.wd.state != 'ready'

		def print_summary(self, signables):
			bmsg('\nAutosign summary:')
			msg('\n'.join(s.get_info(indent='  ') for s in signables) + self.summary_footer)

	class xmr_transaction(xmr_signable):
		dir_name = 'xmr_tx_dir'
		desc = 'Monero transaction'
		subext = 'subtx'
		multiple_ok = False
		summary_footer = ''

		async def sign(self, f):
			from . import xmrwallet
			from .xmrwallet.file.tx import MoneroMMGenTX
			tx1 = MoneroMMGenTX.Completed(self.parent.xmrwallet_cfg, f)
			m = xmrwallet.op(
				'sign',
				self.parent.xmrwallet_cfg,
				infile  = str(self.parent.wallet_files[0]), # MMGen wallet file
				wallets = str(tx1.src_wallet_idx))
			tx2 = await m.main(f, restart_daemon=self.need_daemon_restart(m, tx1.src_wallet_idx))
			tx2.write(ask_write=False)
			return tx2

	class xmr_wallet_outputs_file(xmr_signable):
		desc = 'Monero wallet outputs file'
		rawext = 'raw'
		sigext = 'sig'
		dir_name = 'xmr_outputs_dir'
		clean_all = True
		summary_footer = '\n'

		@property
		def unsigned(self):
			import json
			return tuple(
				f for f in super().unsigned
					if not json.loads(f.read_text())['MoneroMMGenWalletOutputsFile']['data']['imported'])

		async def sign(self, f):
			from . import xmrwallet
			wallet_idx = xmrwallet.op_cls('wallet').get_idx_from_fn(f)
			m = xmrwallet.op(
				'import_outputs',
				self.parent.xmrwallet_cfg,
				infile  = str(self.parent.wallet_files[0]), # MMGen wallet file
				wallets = str(wallet_idx))
			obj = await m.main(f, wallet_idx, restart_daemon=self.need_daemon_restart(m, wallet_idx))
			obj.write(quiet=not obj.data.sign)
			self.action_desc = 'imported and signed' if obj.data.sign else 'imported'
			return obj

	class message(base):
		desc = 'message file'
		rawext = 'rawmsg.json'
		sigext = 'sigmsg.json'
		dir_name = 'msg_dir'
		fail_msg = 'failed to sign or signed incompletely'

		async def sign(self, f):
			from .msg import UnsignedMsg, SignedMsg
			m = UnsignedMsg(self.cfg, infile=f)
			await m.sign(wallet_files=self.parent.wallet_files[:], passwd_file=str(self.parent.keyfile))
			m = SignedMsg(self.cfg, data=m.__dict__)
			m.write_to_file(
				outdir = self.dir.resolve(),
				ask_overwrite = False)
			if m.data.get('failed_sids'):
				die('MsgFileFailedSID', f'Failed Seed IDs: {fmt_list(m.data["failed_sids"], fmt="bare")}')
			return m

		def print_summary(self, signables):
			gmsg('\nSigned message files:')
			for message in signables:
				gmsg('  ' + message.signed_filename)

		def gen_bad_list(self, bad_files):
			for f in bad_files:
				sigfile = f.parent / (f.name[:-len(self.rawext)] + self.sigext)
				yield orange(sigfile.name) if sigfile.exists() else red(f.name)

class Autosign:

	dev_label = 'MMGEN_TX'
	linux_mount_subdir = 'mmgen_autosign'
	macOS_ramdisk_name = 'AutosignRamDisk'
	wallet_subdir = 'autosign'
	linux_blkid_cmd = 'sudo blkid -s LABEL -o value'
	keylist_fn = 'keylist.mmenc'

	cmds = ('setup', 'xmr_setup', 'sign', 'wait')

	util_cmds = (
		'gen_key',
		'macos_ramdisk_setup',
		'macos_ramdisk_delete',
		'enable_swap',
		'disable_swap',
		'clean',
		'wipe_key')

	mn_fmts = {
		'mmgen': 'words',
		'bip39': 'bip39',
	}
	dfl_mn_fmt = 'mmgen'

	non_xmr_dirs = {
		'tx_dir':     'tx',
		'txauto_dir': 'txauto',
		'msg_dir':    'msg',
	}
	xmr_dirs = {
		'xmr_dir':         'xmr',
		'xmr_tx_dir':      'xmr/tx',
		'xmr_outputs_dir': 'xmr/outputs',
	}
	have_xmr = False
	xmr_only = False

	def init_fixup(self): # see test/overlay/fakemods/mmgen/autosign.py
		pass

	def __init__(self, cfg, *, cmd=None):

		if cfg.mnemonic_fmt:
			if cfg.mnemonic_fmt not in self.mn_fmts:
				die(1, '{!r}: invalid mnemonic format (must be one of: {})'.format(
					cfg.mnemonic_fmt,
					fmt_list(self.mn_fmts, fmt='no_spc')))

		if sys.platform == 'linux':
			self.dfl_mountpoint = f'/mnt/{self.linux_mount_subdir}'
			self.dfl_shm_dir    = '/dev/shm'

			# linux-only attrs:
			self.old_dfl_mountpoint = '/mnt/tx'
			self.old_dfl_mountpoint_errmsg = f"""
				Mountpoint ‘{self.old_dfl_mountpoint}’ is no longer supported!
				Please rename ‘{self.old_dfl_mountpoint}’ to ‘{self.dfl_mountpoint}’
				and update your fstab accordingly.
			"""
			self.mountpoint_errmsg_fs = """
				Mountpoint ‘{}’ does not exist or does not point
				to a directory!  Please create the mountpoint and add an entry
				to your fstab as described in this script’s help text.
			"""
		elif sys.platform == 'darwin':
			self.dfl_mountpoint = f'/Volumes/{self.dev_label}'
			self.dfl_shm_dir    = f'/Volumes/{self.macOS_ramdisk_name}'

		self.cfg = cfg

		self.dfl_wallet_dir = f'{self.dfl_shm_dir}/{self.wallet_subdir}'
		self.mountpoint = Path(cfg.mountpoint or self.dfl_mountpoint)
		self.shm_dir    = Path(self.dfl_shm_dir)
		self.wallet_dir = Path(cfg.wallet_dir or self.dfl_wallet_dir)

		if sys.platform == 'linux':
			self.mount_cmd  = f'mount {self.mountpoint}'
			self.umount_cmd = f'umount {self.mountpoint}'
		elif sys.platform == 'darwin':
			self.mount_cmd  = f'diskutil mount {self.dev_label}'
			self.umount_cmd = f'diskutil eject {self.dev_label}'

		self.init_fixup()

		if sys.platform == 'darwin': # test suite uses ‘fixed-up’ shm_dir
			from .platform.darwin.util import MacOSRamDisk
			self.ramdisk = MacOSRamDisk(
					cfg,
					self.macOS_ramdisk_name,
					self._get_macOS_ramdisk_size(),
					path = self.shm_dir)

		self.keyfile = self.mountpoint / 'autosign.key'

		if any(k in cfg._uopts for k in ('help', 'longhelp')):
			return

		self.coins = cfg.coins.upper().split(',') if cfg.coins else []

		if cfg.xmrwallets and not 'XMR' in self.coins:
			self.coins.append('XMR')

		if not self.coins and cmd in self.cmds:
			ymsg('Warning: no coins specified, defaulting to BTC')
			self.coins = ['BTC']

		if 'XMR' in self.coins:
			self.have_xmr = True
			if len(self.coins) == 1:
				self.xmr_only = True
			self.xmr_cur_wallet_idx = None

		self.dirs = {}
		self.signables = ()

		if not self.xmr_only:
			self.dirs |= self.non_xmr_dirs
			self.signables += Signable.non_xmr_signables

		if self.have_xmr:
			self.dirs |= self.xmr_dirs
			self.signables += Signable.xmr_signables

		for name, path in self.dirs.items():
			setattr(self, name, self.mountpoint / path)

		self.swap = SwapMgr(self.cfg, ignore_zram=True)

	async def check_daemons_running(self):
		from .protocol import init_proto
		for coin in self.coins:
			proto = init_proto(self.cfg,  coin, network=self.cfg.network, need_amt=True)
			if proto.sign_mode == 'daemon':
				self.cfg._util.vmsg(f'Checking {coin} daemon')
				from .rpc import rpc_init
				from .exception import SocketError
				try:
					await rpc_init(self.cfg, proto, ignore_wallet=True)
				except SocketError as e:
					from .daemon import CoinDaemon
					d = CoinDaemon(self.cfg, proto=proto, test_suite=self.cfg.test_suite)
					die(2,
						f'\n{e}\nIs the {d.coind_name} daemon ({d.exec_fn}) running '
						+ 'and listening on the correct port?')

	@property
	def wallet_files(self):

		if not hasattr(self, '_wallet_files'):

			try:
				dirlist = self.wallet_dir.iterdir()
			except:
				die(1, f"Cannot open wallet directory '{self.wallet_dir}'. Did you run ‘mmgen-autosign setup’?")

			self._wallet_files = [f for f in dirlist if f.suffix == '.mmdat']

			if not self._wallet_files:
				die(1, 'No wallet files present!')

		return self._wallet_files

	def do_mount(self, *, silent=False, verbose=False):

		def check_or_create(dirname):
			path = getattr(self, dirname)
			if path.is_dir():
				if not path.stat().st_mode & S_IWUSR|S_IRUSR == S_IWUSR|S_IRUSR:
					die(1, f'‘{path}’ is not read/write for this user!')
			elif path.exists():
				die(1, f'‘{path}’ is not a directory!')
			elif path.is_symlink():
				die(1, f'‘{path}’ is a symlink not pointing to a directory!')
			else:
				msg(f'Creating ‘{path}’')
				path.mkdir(parents=True)

		if sys.platform == 'linux' and not self.mountpoint.is_dir():
			def do_die(m):
				die(1, '\n' + yellow(fmt(m.strip(), indent='  ')))
			if Path(self.old_dfl_mountpoint).is_dir():
				do_die(self.old_dfl_mountpoint_errmsg)
			else:
				do_die(self.mountpoint_errmsg_fs.format(self.mountpoint))

		if not self.mountpoint.is_mount():
			redir = None if verbose else DEVNULL
			if run(self.mount_cmd.split(), stderr=redir, stdout=redir).returncode == 0:
				if not silent:
					msg(f"Mounting '{self.mountpoint}'")
			else:
				die(1, f'Unable to mount device {self.dev_label} at {self.mountpoint}')

		for dirname in self.dirs:
			check_or_create(dirname)

	def do_umount(self, *, silent=False, verbose=False):
		if self.mountpoint.is_mount():
			run(['sync'], check=True)
			if not silent:
				msg(f"Unmounting '{self.mountpoint}'")
			redir = None if verbose else DEVNULL
			run(self.umount_cmd.split(), stdout=redir, check=True)
		if not silent:
			bmsg('It is now safe to extract the removable device')

	def decrypt_wallets(self):
		msg(f"Unlocking wallet{suf(self.wallet_files)} with key from ‘{self.keyfile}’")
		fails = 0
		for wf in self.wallet_files:
			try:
				Wallet(self.cfg, fn=wf, ignore_in_fmt=True, passwd_file=str(self.keyfile))
			except SystemExit as e:
				if e.code != 0:
					fails += 1

		return not fails

	async def sign_all(self, target_name):
		target = getattr(Signable, target_name)(self)
		if target.unsigned:
			good = []
			bad = []
			if len(target.unsigned) > 1 and not target.multiple_ok:
				ymsg(f'Autosign error: only one unsigned {target.desc} transaction allowed at a time!')
				target.print_bad_list(target.unsigned)
				return False
			for f in target.unsigned:
				ret = None
				try:
					ret = await target.sign(f)
				except Exception as e:
					ymsg(f"An error occurred with {target.desc} '{f.name}':\n    {type(e).__name__}: {e!s}")
				except:
					ymsg(f"An error occurred with {target.desc} '{f.name}'")
				good.append(ret) if ret else bad.append(f)
				self.cfg._util.qmsg('')
			await asyncio.sleep(0.3)
			msg(brown(f'{len(good)} {target.desc}{suf(good)} {target.action_desc}'))
			if bad:
				rmsg(f'{len(bad)} {target.desc}{suf(bad)} {target.fail_msg}')
			if good and not self.cfg.no_summary:
				target.print_summary(good)
			if bad:
				target.print_bad_list(bad)
			return not bad
		else:
			return f'No unsigned {target.desc}s'

	async def do_sign(self):
		if not self.cfg.stealth_led:
			self.led.set('busy')
		self.do_mount()
		key_ok = self.decrypt_wallets()
		self.init_non_mmgen_keys()
		if key_ok:
			if self.cfg.stealth_led:
				self.led.set('busy')
			ret = [await self.sign_all(signable) for signable in self.signables]
			for val in ret:
				if isinstance(val, str):
					msg(val)
			if self.cfg.test_suite_autosign_threaded:
				await asyncio.sleep(0.3)
			self.do_umount()
			self.led.set('error' if not all(ret) else 'off' if self.cfg.stealth_led else 'standby')
			return all(ret)
		else:
			msg('Password is incorrect!')
			self.do_umount()
			if not self.cfg.stealth_led:
				self.led.set('error')
			return False

	def wipe_encryption_key(self):
		if self.keyfile.exists():
			ymsg(f'Shredding wallet encryption key ‘{self.keyfile}’')
			shred_file(self.cfg, self.keyfile)
		else:
			gmsg('No wallet encryption key on removable device')

	def create_key(self):
		desc = f"key file '{self.keyfile}'"
		msg('Creating ' + desc)
		try:
			self.keyfile.write_text(os.urandom(32).hex())
			self.keyfile.chmod(0o400)
		except:
			die(2, 'Unable to write ' + desc)
		msg('Wrote ' + desc)

	def gen_key(self, *, no_unmount=False):
		if not self.device_inserted:
			die(1, 'Removable device not present!')
		self.do_mount()
		self.wipe_encryption_key()
		self.create_key()
		if not no_unmount:
			self.do_umount()

	def macos_ramdisk_setup(self):
		self.ramdisk.create()

	def macos_ramdisk_delete(self):
		self.ramdisk.destroy()

	def _get_macOS_ramdisk_size(self):
		from .platform.darwin.util import MacOSRamDisk, warn_ramdisk_too_small
		# allow 1MB for each Monero wallet
		xmr_size = len(AddrIdxList(fmt_str=self.cfg.xmrwallets)) if self.cfg.xmrwallets else 0
		calc_size = xmr_size + 1
		usr_size = self.cfg.macos_ramdisk_size or self.cfg.macos_autosign_ramdisk_size
		if is_int(usr_size):
			usr_size = int(usr_size)
		else:
			die(1, f'{usr_size}: invalid user-specified macOS ramdisk size (not an integer)')
		min_size = MacOSRamDisk.min_size
		size = max(usr_size, calc_size, min_size)
		if usr_size and usr_size < min_size:
			warn_ramdisk_too_small(usr_size, min_size)
		return size

	def setup(self):

		def remove_wallet_dir():
			msg(f"Deleting '{self.wallet_dir}'")
			import shutil
			try:
				shutil.rmtree(self.wallet_dir)
			except:
				pass

		def create_wallet_dir():
			try:
				self.wallet_dir.mkdir(parents=True)
			except:
				pass
			try:
				self.wallet_dir.stat()
			except:
				die(2, f"Unable to create wallet directory '{self.wallet_dir}'")

		self.gen_key(no_unmount=True)

		self.swap.disable()

		if sys.platform == 'darwin':
			self.macos_ramdisk_setup()

		remove_wallet_dir()
		create_wallet_dir()

		wf = find_file_in_dir(get_wallet_cls('mmgen'), self.cfg.data_dir)
		if wf and keypress_confirm(
				cfg         = self.cfg,
				prompt      = f"Default wallet '{wf}' found.\nUse default wallet for autosigning?",
				default_yes = True):
			ss_in = Wallet(Config(), fn=wf)
		else:
			ss_in = Wallet(self.cfg, in_fmt=self.mn_fmts[self.cfg.mnemonic_fmt or self.dfl_mn_fmt])
		ss_out = Wallet(self.cfg, ss=ss_in, passwd_file=str(self.keyfile))
		ss_out.write_to_file(desc='autosign wallet', outdir=self.wallet_dir)

		if self.cfg.keys_from_file:
			self.setup_non_mmgen_keys()

	@property
	def xmrwallet_cfg(self):
		if not hasattr(self, '_xmrwallet_cfg'):
			self._xmrwallet_cfg = Config({
				'_clone': self.cfg,
				'coin': 'xmr',
				'wallet_rpc_user': 'autosign',
				'wallet_rpc_password': 'autosign password',
				'wallet_rpc_port': 23232 if self.cfg.test_suite_xmr_autosign else None,
				'wallet_dir': str(self.wallet_dir),
				'autosign': True,
				'autosign_mountpoint': str(self.mountpoint),
				'offline': True,
				'passwd_file': str(self.keyfile),
			})
		return self._xmrwallet_cfg

	def xmr_setup(self):

		def create_signing_wallets():
			from . import xmrwallet
			if len(self.wallet_files) > 1:
				ymsg(f'Warning: more than one wallet file, using the first ({self.wallet_files[0]}) for xmrwallet generation')
			m = xmrwallet.op(
				'create_offline',
				self.xmrwallet_cfg,
				infile  = str(self.wallet_files[0]), # MMGen wallet file
				wallets = self.cfg.xmrwallets)       # XMR wallet idxs
			asyncio.run(m.main())
			asyncio.run(m.stop_wallet_daemon())

		self.clean_old_files()

		create_signing_wallets()

	def clean_old_files(self):

		def do_shred(fn):
			nonlocal count
			msg_r('.')
			shred_file(self.cfg, fn, iterations=15)
			count += 1

		def clean_dir(s_name):

			def clean_files(rawext, sigext):
				for f in s.dir.iterdir():
					if s.clean_all and (f.name.endswith(f'.{rawext}') or f.name.endswith(f'.{sigext}')):
						do_shred(f)
					elif f.name.endswith(f'.{sigext}'):
						raw = f.parent / (f.name[:-len(sigext)] + rawext)
						if raw.is_file():
							do_shred(raw)

			s = getattr(Signable, s_name)(self)

			msg_r(f"Cleaning directory '{s.dir}'..")

			if s.dir.is_dir():
				clean_files(s.rawext, s.sigext)
				if hasattr(s, 'subext'):
					clean_files(s.rawext, s.subext)
					clean_files(s.sigext, s.subext)

			msg('done' if s.dir.is_dir() else 'skipped (no dir)')

		count = 0

		for s_name in self.signables:
			clean_dir(s_name)

		bmsg(f'{count} file{suf(count)} shredded')

	@property
	def device_inserted(self):
		if self.cfg.no_insert_check:
			return True
		if sys.platform == 'linux':
			cp = run(self.linux_blkid_cmd.split(), stdout=PIPE, text=True)
			if cp.returncode not in (0, 2):
				die(2, f'blkid exited with error code {cp.returncode}')
			return self.dev_label in cp.stdout.splitlines()
		elif sys.platform == 'darwin':
			if self.cfg.test_suite_root_pfx:
				return self.mountpoint.exists()
			else:
				return run(['diskutil', 'info', self.dev_label], stdout=DEVNULL, stderr=DEVNULL).returncode == 0

	async def main_loop(self):
		if not self.cfg.stealth_led:
			self.led.set('standby')
		threaded = self.cfg.test_suite_autosign_threaded
		n = 1 if threaded else 0
		prev_status = False
		while True:
			status = self.device_inserted
			if status and not prev_status:
				msg('Device insertion detected')
				await self.do_sign()
			prev_status = status
			if not n % 10:
				msg_r(f"\r{' '*17}\rWaiting")
			await asyncio.sleep(0.2 if threaded else 1)
			if not threaded:
				msg_r('.')
				n += 1

	def at_exit(self, exit_val, message=None):
		if message:
			msg(message)
		self.led.stop()
		sys.exit(0 if self.cfg.test_suite_autosign_threaded else int(exit_val))

	def init_exit_handler(self):

		def handler(arg1, arg2):
			self.at_exit(1, '\nCleaning up...')

		import signal
		signal.signal(signal.SIGTERM, handler)
		signal.signal(signal.SIGINT, handler)

	def init_led(self):
		from .led import LEDControl
		self.led = LEDControl(
			enabled = self.cfg.led,
			simulate = self.cfg.test_suite_autosign_led_simulate)
		self.led.set('off')

	def setup_non_mmgen_keys(self):
		from .fileutil import get_lines_from_file, write_data_to_file
		from .crypto import Crypto
		lines = get_lines_from_file(self.cfg, self.cfg.keys_from_file, desc='keylist data')
		write_data_to_file(
			self.cfg,
			str(self.wallet_dir / self.keylist_fn),
			Crypto(self.cfg).mmgen_encrypt(
				data = '\n'.join(lines).encode(),
				passwd = self.keyfile.read_text()),
			desc = 'encrypted keylist data',
			binary = True)
		if keypress_confirm(self.cfg, 'Securely delete original keylist file?'):
			shred_file(self.cfg, self.cfg.keys_from_file)

	def init_non_mmgen_keys(self):
		if not hasattr(self, 'keylist'):
			path = self.wallet_dir / self.keylist_fn
			if path.exists():
				from .crypto import Crypto
				from .fileutil import get_data_from_file
				self.keylist = Crypto(self.cfg).mmgen_decrypt(
					get_data_from_file(
						self.cfg,
						path,
						desc = 'encrypted keylist data',
						binary = True),
					passwd = self.keyfile.read_text()).decode().split()
			else:
				self.keylist = None



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.__init__: wallet class initializer
"""

import importlib
from collections import namedtuple

from ..util import die, get_extension
from ..objmethods import MMGenObject
from ..seed import Seed

_wd = namedtuple('wallet_data', ['type', 'name', 'ext', 'base_type', 'enc', 'fmt_codes'])

wallet_data = {
	'bip39':
_wd('bip39',        'BIP39Mnemonic',     'bip39',   'mnemonic',   False, ('bip39',)),
	'brain':
_wd('brain',        'Brainwallet',       'mmbrain', None,         True,  ('mmbrain','brainwallet','brain','bw')),
	'dieroll':
_wd('dieroll',      'DieRollWallet',     'b6d',     None,         False, ('b6d','die','dieroll')),
	'incog':
_wd('incog',        'IncogWallet',       'mmincog', 'incog_base', True,  ('mmincog','incog','icg','i')),
	'incog_hex':
_wd('incog_hex',    'IncogWalletHex',    'mmincox', 'incog_base', True,  ('mmincox','incox','incog_hex','ix','xi')),
	'incog_hidden':
_wd('incog_hidden', 'IncogWalletHidden', None,      'incog_base', True,  ('incog_hidden','hincog','ih','hi')),
	'mmgen':
_wd('mmgen',        'MMGenWallet',       'mmdat',   None,         True,  ('wallet','w')),
	'mmhex':
_wd('mmhex',        'MMGenHexSeedFile',  'mmhex',   None,         False, ('seedhex','hexseed','mmhex')),
	'plainhex':
_wd('plainhex',     'PlainHexSeedFile',  'hex',     None,         False, ('hex','rawhex','plainhex')),
	'seed':
_wd('seed',         'MMGenSeedFile',     'mmseed',  None,         False, ('mmseed','seed','s')),
	'words':
_wd('words',        'MMGenMnemonic',     'mmwords', 'mnemonic',   False, ('mmwords','words','mnemonic','mn','m')),
}

def get_wallet_data(
		*,
		wtype       = None,
		fmt_code    = None,
		ext         = None,
		die_on_fail = False):

	if wtype:
		return wallet_data[wtype]
	elif fmt_code:
		for v in wallet_data.values():
			if fmt_code in v.fmt_codes:
				return v
	elif ext is not None: # ext could be the empty string
		for v in wallet_data.values():
			if ext == v.ext:
				return v
	else:
		die(4, 'no argument supplied!')

	if die_on_fail:
		if fmt_code:
			die(3, f'{fmt_code!r}: unrecognized wallet format code')
		else:
			die('BadFileExtension', f'{ext!r}: unrecognized wallet file extension')

def get_wallet_cls(
		wtype       = None,
		*,
		fmt_code    = None,
		ext         = None,
		die_on_fail = False):

	return getattr(
		importlib.import_module('mmgen.wallet.{}'.format(
			wtype or
			get_wallet_data(
				fmt_code    = fmt_code,
				ext         = ext,
				die_on_fail = die_on_fail).type
		)),
		'wallet')

def get_wallet_extensions(key):
	return {
		'enc':   [v.ext for v in wallet_data.values() if v.enc],
		'unenc': [v.ext for v in wallet_data.values() if not v.enc]
	}[key]

def format_fmt_codes():
	d = [(
			v.name,
			('.' + v.ext if v.ext else 'None'),
			','.join(v.fmt_codes)
		) for v in wallet_data.values()]
	w = max(len(i[0]) for i in d)
	ret = [f'{a:<{w}}  {b:<9} {c}' for a, b, c in [
		('Format', 'FileExt', 'Valid codes'),
		('------', '-------', '-----------')
		] + sorted(d)]
	return '\n'.join(ret) + '\n'

def _get_me(modname):
	return MMGenObject.__new__(getattr(importlib.import_module(f'mmgen.wallet.{modname}'), 'wallet'))

def Wallet(
	cfg,
	*,
	fn            = None,
	ss            = None,
	seed_bin      = None,
	seed          = None,
	passchg       = False,
	in_data       = None,
	ignore_in_fmt = False,
	in_fmt        = None,
	passwd_file   = None):

	in_fmt = in_fmt or cfg.in_fmt

	ss_out = (
		get_wallet_data(
			fmt_code    = cfg.out_fmt,
			die_on_fail = True).type
		if cfg.out_fmt else None)

	if seed or seed_bin:
		me = _get_me(ss_out or 'mmgen') # default to native wallet format
		me.seed = seed or Seed(cfg, seed_bin=seed_bin)
		me.op = 'new'
	elif ss:
		me = _get_me(ss.type if passchg else (ss_out or 'mmgen'))
		me.seed = ss.seed
		me.ss_in = ss
		me.op = 'pwchg_new' if passchg else 'conv'
	elif fn or cfg.hidden_incog_input_params:
		if fn:
			wd = get_wallet_data(ext=get_extension(fn), die_on_fail=True)
			if in_fmt and (not ignore_in_fmt) and in_fmt not in wd.fmt_codes:
				die(1, f'{in_fmt}: --in-fmt parameter does not match extension of input file')
			me = _get_me(wd.type)
		else:
			fn = ','.join(cfg.hidden_incog_input_params.split(',')[:-1]) # permit comma in filename
			me = _get_me('incog_hidden')
		from ..filename import MMGenFile
		me.infile = MMGenFile(fn, subclass=type(me))
		me.op = 'pwchg_old' if passchg else 'old'
	elif in_fmt:
		me = _get_me(get_wallet_data(fmt_code=in_fmt).type)
		me.op = 'pwchg_old' if passchg else 'old'
	else: # called with no arguments: initialize with random seed
		me = _get_me(ss_out or 'mmgen') # default to native wallet format
		me.seed = Seed(cfg)
		me.op = 'new'

	me.cfg = cfg

	me.__init__(
		in_data     = in_data,
		passwd_file = passwd_file)

	return me

def check_wallet_extension(fn):
	get_wallet_data(ext=get_extension(fn), die_on_fail=True) # raises exception on failure



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
swap.asset: swap asset class for the MMGen Wallet suite
"""

from collections import namedtuple

from ..util import die

class SwapAsset:

	_ad = namedtuple('swap_asset_data', ['desc', 'name', 'full_name', 'abbr', 'tested'])
	assets_data = {}
	evm_contracts = {}
	unsupported = ()
	blacklisted = {}
	evm_chains = ()

	def fmt_assets_data(self, indent=''):

		def gen_good():
			fs = '%s{:10} {:23} {:9} {}' % indent
			yield fs.format('ASSET', 'DESCRIPTION', 'STATUS', 'CONTRACT ADDRESS')
			for k, v in self.assets_data.items():
				if not k in self.blacklisted:
					if k in self.send or k in self.recv:
						yield fs.format(
							k,
							v.desc,
							'tested' if v.tested else 'untested',
							self.evm_contracts.get(k,'-'))

		def gen_bad():
			if self.blacklisted:
				fs = '%s{:10} {:23} {}' % indent
				yield '\n\nBlacklisted assets:'
				yield fs.format('ASSET', 'DESCRIPTION', 'REASON')
				for k, v in self.blacklisted.items():
					yield fs.format(k, self.assets_data[k].desc, v)

		return '\n'.join(gen_good()) + '\n'.join(gen_bad())

	@classmethod
	def init_from_memo(cls, s):
		for d in cls.assets_data.values():
			if s in (d.abbr, d.full_name):
				return cls(d.name or d.full_name, 'recv')
		die('SwapAssetError', f'{s!r}: unrecognized asset name or abbreviation')

	@property
	def tested(self):
		return [k for k, v in self.assets_data.items() if v.tested]

	@property
	def send(self):
		return set(self.assets_data) - set(self.unsupported) - set(self.blacklisted)

	@property
	def recv(self):
		return set(self.assets_data) - set(self.unsupported) - set(self.blacklisted)

	@property
	def chain(self):
		return self.data.full_name.split('.', 1)[0] if self.data.full_name else self.name

	@property
	def coin(self):
		return self.data.name or self.data.full_name.split('.', 1)[0]

	@property
	def full_name(self):
		return self.data.full_name or f'{self.data.name}.{self.data.name}'

	@property
	def short_name(self):
		return self.data.name or self.data.full_name.split('.', 1)[1]

	@property
	def tokensym(self):
		return None if self.data.name else self.data.full_name.split('.', 1)[1]

	@property
	def memo_asset_name(self):
		return self.data.abbr or self.data.full_name

	def __init__(self, name, direction):

		if name not in self.assets_data:
			die('SwapAssetError', f'{name!r}: unrecognized asset')

		assert direction in ('send', 'recv'), 'direction must be ‘send’ or ‘recv’'

		if direction == 'send' and name not in self.send:
			die('SwapAssetError', f'{name!r} unsupported send asset')

		if direction == 'recv' and name not in self.recv:
			die('SwapAssetError', f'{name!r} unsupported receive asset')

		self.direction = direction
		self.name = name
		self.data = self.assets_data[name]
		self.desc = self.data.desc
